'''
colors.py
Some functions for color conversion and a definition of the canonical color
system. These convert normalized h, s, and v values (on [0,1]) into integer H,
S, and V values (on [0,11], [0,4], and [0,9] respectively). There is also a
conversion function from a color name to a normalized hue value.
'''

# A list for lookups:
colors = [
  "red",
  "orange",
  "yellow",
  "lime",
  "green",
  "teal",
  "cyan",
  "azure",
  "blue",
  "purple",
  "magenta",
  "pink",
]

# Convenient module variables:
red = 0
orange = 1
yellow = 2
lime = 3
green = 4
teal = 5
cyan = 6
azure = 7
blue = 8
purple = 9
magenta = 10
pink = 11

# A map for conversion to and from normalized hues. Each value contains the
# corresonding normalized midpoint and upper endpoint of the hue range for that
# color name.
colormap = {
  "red":       (  0/360.,  15/360.),
  "orange":    ( 30/360.,  50/360.),
  "yellow":    ( 60/360.,  67/360.),
  "lime":      ( 75/360.,  83/360.),
  "green":     (120/360., 145/360.),
  "teal":      (163/360., 171/360.),
  "cyan":      (180/360., 190/360.),
  "azure":     (210/360., 225/360.),
  "blue":      (240/360., 273/360.),
  "purple":    (280/360., 290/360.),
  "magenta":   (300/360., 310/360.),
  "pink":      (320/360., 335/360.),
}

# Misc functions:

def prevH(H):
  prev = H - 1
  if prev < 0:
    prev = len(colors) - 1
  return prev

def nextH(H):
  next = H + 1
  if next >= len(colors):
    next = 0
  return next

def huelow(hue):
  return colormap[Hhue(prevH(hueH(hue)))][1] + 1/360.

def huemid(hue):
  return colormap[hue][0]

def huehigh(hue):
  return colormap[hue][1]

# Conversion functions:

def hH(hue):
  idx = 0
  while idx < len(colors) and (hue > colormap[colors[idx]][1]):
    idx += 1
  if idx >= len(colors):
    idx = 0
  return idx

def Hh(hue):
  if hue >= len(colormap):
    return 0.0
  return colormap[colors[hue]][0]

def hhue(hue):
  return colors[hH(hue)]

def hueh(hue):
  return colormap[hue][0]

def Hhue(hue):
  return colors[hue]

def hueH(hue):
  return colors.index(hue)

def sS(sat):
  if sat > 1.0:
    return 4
  return max(0, (int(sat*100) - 10)//20) # 30 to 109 by 20s <=> 1-4

def Ss(sat):
  if sat <= 0:
    return 0
  elif sat > 4:
    return 1.0
  return (20 + sat*20)/100 # 40 to 100 by 20s <=> 1-4

def vV(val):
  if val > 1.0:
    return 9
  return max(0, (int(val*100) - 24)//8) # 32 to 103 by 8s <=> 1-9

def Vv(val):
  if val <= 0:
    return 0
  elif val > 9:
    return 1.0
  return (24 + val*8 + 4)/100 # 36 to 100 by 8s <=> 1-9

def vu(val):
  return max(0, (int(val*100) - 4)// 6) # 10 to 105 by 6s <=> 1-16

def uv(val):
  if val <= 0:
    return 0
  elif val > 15:
    return 1.0
  return (4 + val*6 + 3)/100 # 13 to 97 <=> 1-15

def hsvHSV(h, s, v):
  return hH(h), sS(s), vV(v)

def HSVhsv(H, S, V):
  return Hh(H), Ss(S), Vv(V)

def Ve(value):
  '''
  Converts from Value to energy. Takes a Value number (1-9) and returns an
  energy number (0-32). Giving a 0 Value results in -4 energy.
  '''
  return (value - 1)*4

def eV(energy):
  '''
  Converts from energy to Value. The inverse of the Ve function. Giving *any*
  negative energy returns a 0 Value.
  '''
  if energy < 0:
    return 0
  return (energy // 4) + 1

class Color:
  '''
  A Color stores h, s, and v normalized color values, and makes them accessible
  as H, S and V integer values and a hue name according to the mappings defined
  in this module.
  '''
  h = None
  s = None
  v = None
  def __init__(self, h=None, s=None, v=None, H=None, S=None, V=None, hue=None):
    '''
    Sets up the pixel with hue, sautration, and value arguments. They should
    each be in [0,1]. Alternatively, H, S, or V integer arguments, or a 'hue'
    color name argument may be given. The two argument sets may be mixed, and
    if for some reason both are given, the normalized versions override their
    less general counterparts.
    '''
    if hue:
      self.hue = hue
    if H:
      self.H = H
    if S:
      self.S = S
    if V:
      self.V = V
    if h != None:
      self.h = h
    elif self.h == None:
      self.h = 0
    if s != None:
      self.s = s
    elif self.s == None:
      self.s = 0
    if v != None:
      self.v = v
    elif self.v == None:
      self.v = 0

  @staticmethod
  def parse(string):
    '''
    Parses a color definition out of a string. Colors can be defined using
    %h,s,v or ~H,S,V syntax.
    '''
    if string[0] == '%':
      h, s, v = string[1:].split(',')
      h, s, v = (float(val.strip()) for val in (h, s, v))
      return Color(h, s, v)
    elif string[0] == '~':
      H, S, V = string[1:].split(',')
      H, S, V = (int(val.strip()) for val in (H, S, V))
      return Color(H=H, S=S, V=V)
    else:
      raise ValueError("Bad color string (should start with '%' or '~').")

  def matches(self, other):
    '''
    Checks that this color matches another, using the more permissive H, S, V
    scales.
    '''
    if not isinstance(other, Color):
      raise TypeError("Can't match against a non-Color.")
    return self.H == other.H and self.S == other.S and self.V == other.V

  @property
  def hue(self):
    '''
    Read or write the h component of the hue as a color name.
    '''
    return hhue(self.h)

  @hue.setter
  def hue(self, value):
    self.h = hueh(value)

  @property
  def H(self):
    '''
    Read or write the h component as an integer.
    '''
    return hH(self.h)

  @H.setter
  def H(self, value):
    self.h = Hh(value)

  @property
  def S(self):
    '''
    Read or write the s component as an integer.
    '''
    return sS(self.s)

  @S.setter
  def S(self, value):
    self.s = Ss(value)

  @property
  def V(self):
    '''
    Read or write the v component as an integer.
    '''
    return vV(self.v)

  @V.setter
  def V(self, value):
    self.v = Vv(value)
