'''
engine.py
Handles applying rules to a world and tracking signals between steps.
'''

import collections

from types import FunctionType as function

from . import colors, frwrap

from . import debug

###########
# Globals #
###########

VALUES = {}
POSITIONS = {}
SIGNALS = set()
NEXT_SIGNALS = set()

UNIVERSE = {}

STEP = 0

NS_BIAS = 'c'
EW_BIAS = 'c'

# TILT controls how much edge overlap is required in order to move between
# fraxels on neighboring embedded surfaces. If the denominator isn't larger
# than the largest surface dimension, it will cause problems when moving
# between two adjacent embedded surfaces (movement may not be possible).
TILT = 1/100000

####################
# Helper functions #
####################

def make_unknown(surface):
  '''
  Performs operations on a surface to set up the canonical "unknown" surface.
  The given surface should be 5x5 and have scale dimensions of 1x1.
  '''
  grey = frwrap.Pixel(0, 0, V=8)
  black = frwrap.Pixel(0, 0, 0)
  ref = frwrap.Reference(surface, 'n')

  surface.fill(grey)

  surface.set(0, 2, black)
  surface.set(2, 0, black)
  surface.set(4, 2, black)
  surface.set(2, 4, black)
  surface.set(2, 2, black)

  surface.set(2, 1, ref)
  surface.set(1, 2, ref)
  surface.set(2, 3, ref)
  surface.set(3, 2, ref)

def get_step():
  return STEP

def set_biases(ns, ew):
  global NS_BIAS, EW_BIAS
  NS_BIAS = ns
  EW_BIAS = ew

def set_tilt(tilt):
  global TILT
  TILT = tilt

def dirint(dir):
  if dir == 'n':
    return 0
  elif dir == 'e':
    return 1
  elif dir == 's':
    return 2
  elif dir == 'w':
    return 3
  return None

def dirchar(dir):
  dir %= 4
  if dir == 0:
    return 'n'
  elif dir == 1:
    return 'e'
  elif dir == 2:
    return 's'
  elif dir == 3:
    return 'w'

def get_value(name):
  if name in VALUES:
    return VALUES[name]
  return None

def set_value(name, value):
  VALUES[name] = value

def clear_values():
  global VALUES
  VALUES = {}

def get_position(name):
  if name in POSITIONS:
    return POSITIONS[name]
  return None

def get_all_positions():
  return POSITIONS.values()

def set_position(name, value):
  debug.debug("positions", "Setting position '{}':".format(name), value)
  if not isinstance(value, frwrap.Position):
    raise TypeError("Can't set a Position to some other kind of value.")
  POSITIONS[name] = value

def clear_positions():
  global POSITIONS
  POSITIONS = {}

def check_signals(*names):
  return all(name in SIGNALS for name in names)

def get_signals(prefix):
  return set([s for s in SIGNALS if s.startswith(prefix)])

def raise_signals(*names):
  NEXT_SIGNALS.update(names)

def clear_signals():
  global NEXT_SIGNALS
  NEXT_SIGNALS = set()

def set_universe(universe):
  global UNIVERSE
  UNIVERSE = universe

def get_world(name):
  '''
  Looks up a world by name within the current universe. Returns a world (and
  OrderedDict object) or None.
  '''
  if name in UNIVERSE:
    return UNIVERSE[name]
  return None

def make_world(name):
  '''
  Creates a new empty world and returns it.
  '''
  frwrap.check_name(name)
  world = collections.OrderedDict()
  world['<name>'] = name

  # Create and add the 'unknown' surface:
  UNKNOWN = frwrap.Surface("<unknown>", world, 1, 1, 5, 5)
  make_unknown(UNKNOWN)

  return world

def add_world(world):
  '''
  Adds the given world to the current universe. Note that unlike Surfaces,
  which automatically add themselves to their worlds, worlds are not
  automatically added to any universe when loaded or created.
  '''
  global UNIVERSE
  if not isinstance(world, collections.OrderedDict):
    raise TypeError("World must be an ordered dictionary.")
  frwrap.check_name(world['<name>'])
  UNIVERSE[world['<name>']] = world

def reset():
  '''
  Resets the world state by clearing values, positions, and signals.
  '''
  clear_values()
  clear_positions()
  clear_signals()

def find_surface(name, world=None):
  '''
  Looks up a surface by name, either within the given world, or using the
  dotted world.surface name syntax. Returns a surface or None.
  '''
  if world:
    if name in world:
      return world[name]
    return None
  else:
    consider = []
    if '.' in name:
      wn, *rest = name.split('.')
      sn = '.'.join(rest)
      if wn in UNIVERSE:
        consider.append(UNIVERSE[wn])
        name = sn
    if not consider:
      consider = UNIVERSE.keys()
    for wn in consider:
      if name in UNIVERSE[wn]:
        return UNIVERSE[wn][name]
    return None

def move(frm, to, clear=None):
  '''
  Takes two Fraxels and sets the value of the second to the value of the first.
  It then replaces the first with the given clear value (if one was given). If
  the thing moved is a reference and the old reference is cleared, it then
  calls fix_positions.
  '''
  if not isinstance(clear, colors.Color)\
  and not isinstance(clear, frwrap.Reference):
    raise TypeError("Replacement must be a Color, Pixel, or Reference.")
  to.become(frm)
  if clear:
    frm.become(clear)
    fix_positions(frm, to)

def fix_positions(frm, to):
  '''
  Iterates over all stored positions and moves any that point to the Fraxel frm
  to point to the Fraxel to instead. frm and to must be on the same Surface.
  '''
  if frm.surface != to.surface:
    raise ValueError("Source and destination must be on the same Surface.")
  seen = set()
  for p in get_all_positions():
    current = p
    debug.debug("fix_positions", "Fixing:", p)
    while current != None and id(current) not in seen:
      debug.debug("fix_positions", "Considering:", current)
      if current.surface == frm.surface\
      and current.x == frm.x\
      and current.y == frm.y:
        debug.debug("fix_positions", "  Moving:", current)
        current.x = to.x
        current.y = to.y
        # Surface is already the same...
      seen.add(id(current))
      current = current.parent

def get_neighbor(position, direction, test=lambda fr: True):
  '''
  Returns a Position that neighbors the given position in the given direction.
  The position will always point to either a pixel or an edge fraxel (never to
  a reference). Takes the given test into account, and returns None if there is
  no suitable neighbor in the given direction. The direction should be one of
  'n', 'e', 's', or 'w', but may optionally be a string composed from these
  characters. In that case, the result is the same as calling get_neighbor with
  each of those directions one at a time, and checking for a None result each
  time (i.e. it returns the final position result from moving in all of the
  specified directions one after the other, or None if any of those moves
  fail). In the extra-special case that one or more '|' characters are given
  as part of the directions list, these represent checkpoints: a failed search
  after a '|' character returns the position that was found up to the '|'
  character.
  '''
  global NS_BIAS, EW_BIAS
  if len(direction) > 1:
    current = position
    result = None
    for dir in direction:
      if dir == '|':
        result = current
      else:
        next = get_neighbor(current, dir, test=test)
        if not next:
          return result
        current = next
    result = current
    return result
  # else:
  posfr = position.fraxel
  if direction == 'n':
    neighbor = posfr.north
  if direction == 'e':
    neighbor = posfr.east
  if direction == 's':
    neighbor = posfr.south
  if direction == 'w':
    neighbor = posfr.west
  if neighbor.is_pixel:
    if test(neighbor):
      return frwrap.Position(
        neighbor.x,
        neighbor.y,
        neighbor.surface,
        position.get_parent()
      )
    else:
      return None
  elif neighbor.is_reference:
    enp = entry_point(neighbor, direction, test, biasNS=NS_BIAS, biasEW=EW_BIAS)
    if enp:
      return frwrap.Position(
        enp.x,
        enp.y,
        enp.surface,
        frwrap.Position(
          neighbor.x,
          neighbor.y,
          neighbor.surface,
          position.get_parent()
        )
      )
    else:
      return None
  elif neighbor.is_edge:
    # Find the exit point:
    exp = exit_point(position, neighbor, direction)
    if not exp: # When there's an outermost surface...
      return None
    # The reentrant case:
    if exp.is_reference:
      # Compute min an max:
      mn, mx = 0, 1
      if direction in 'ns':
        x = position.x
        sc = position.surface.hz_scale
        while x >= sc:
          x -= sc
        mn = x / sc
        mx = (x + 1) / sc
      else:
        y = position.y
        sc = position.surface.vt_scale
        while y >= sc:
          y -= sc
        mn = y / sc
        mx = (y + 1) / sc
      # Flip min and max (if necessary):
      orientation = position.get_parent().fraxel.orientation
      if orientation == 'e' and direction in 'ew'\
      or orientation == 's'\
      or orientation == 'w' and direction in 'ns':
        mn, mx = 1 - mx, 1 - mn
      # Compute the new direction of entry:
      newdir = dirint(direction) + dirint(orientation)
      newdir = dirchar(newdir)
      # Flip min and max again (if necessary):
      orientation = exp.orientation
      if orientation == 'e' and newdir in 'ew'\
      or orientation == 's'\
      or orientation == 'w' and newdir in 'ns':
        mn, mx = 1 - mx, 1 - mn
      # Get the entry point:
      enp = entry_point(
        exp, newdir, test,
        mn=mn, mx=mx,
        biasNS=NS_BIAS, biasEW=EW_BIAS
      )
      if enp:
        return frwrap.Position(
          enp.x,
          enp.y,
          enp.surface,
          frwrap.Position(
            exp.x,
            exp.y,
            exp.surface,
            position.get_parent().get_parent()
          )
        )
      else:
        return None
    # The normal exit case:
    else:
      if exp and test(exp):
        return frwrap.Position(
          exp.x,
          exp.y,
          exp.surface,
          position.get_parent().get_parent()
        )
      else:
        return None
  # This shouldn't be possible...
  raise NotImpementedError("Missed all cases in get_neighbor.")

def entry_point(ref, dir, clear_test, mn=0, mx=1, biasNS='c', biasEW='c'):
  '''
  A helper function for move_position that computes the entry point into a
  surface given the particular referent to that surface moved into, the
  direction of movement, and the clear_test. Optionally, a minimum and maximum
  fraction of the moved-into side may be given. These are then used to restrain
  which fraxels can be entered. They should be given as fractions between 0 and
  1, and should always refer to minimum and maximum x or y values respectively.
  Additionally, north-south and east-west biases may be given (they should be
  in ('n', 'c', 's') and ('e', 'c', 'w') respectively). These are used to bias
  the selection among enterable fraxels, and both default to 'c' (centered
  relative to the given mn and mx values).
  '''
  if not ref.is_reference:
    raise TypeError("Can't move into a non-reference.")
  # Determine which side we are entering:
  side = dirint(dir) - 2
  side -= dirint(ref.orientation)
  while side < 0:
    side += 4

  # Find the root of this reference and determine the offsets:
  root = frwrap.Surface.find_root_of(ref)
  offx = ref.x - root.x
  offy = ref.y - root.y

  # Fix the offsets, I guess:
  # TODO: Does this work for non-square surfaces?
  if ref.orientation == 'e':
    offx, offy = offy, ref.referent.sc_height - 1 - offx
  elif ref.orientation == 's':
    offx = ref.referent.sc_width - 1 - offx
    offy = ref.referent.sc_height - 1 - offy
  elif ref.orientation == 'w':
    offx, offy = ref.referent.sc_height - 1 - offy, offx

  # Determine the north-south and east-west bias points:
  if biasNS == 'n':
    biasNS = mn
  elif biasNS == 's':
    biasNS = mx
  elif biasNS == 'c':
    biasNS = mn + (mx - mn) / 2
  if biasEW == 'w':
    biasEW = mn
  elif biasEW == 'e':
    biasEW = mx
  elif biasEW == 'c':
    biasEW = mn + (mx - mn) / 2

  # Turn the side number into a list of Fraxels (which should all be Pixels):
  if side == 0: # north
    sc = ref.referent.hz_scale
    first = offx * sc
    side = [
      ref.referent.get(x, 0)
      for x in sorted(
        range(first, first + sc),
        key=lambda i: abs((i - first + 0.5)/sc - biasEW)
      )
      if (x - first) / sc < mx - TILT and (x + 1 - first) / sc > mn + TILT
    ]
  elif side == 1: # east
    sc = ref.referent.vt_scale
    first = offy * sc
    w = ref.referent.width
    side = [
      ref.referent.get(w - 1, y)
      for y in sorted(
        range(first, first + sc),
        key=lambda i: abs((i - first + 0.5)/sc - biasNS)
      )
      if (y - first) / sc < mx - TILT and (y + 1 - first) / sc > mn + TILT
    ]
  elif side == 2: # south
    sc = ref.referent.hz_scale
    first = offx * sc
    h = ref.referent.height
    side = [
      ref.referent.get(x, h - 1)
      for x in sorted(
        range(first, first + sc),
        key=lambda i: abs((i - first + 0.5)/sc - biasEW)
      )
      if (x - first) / sc < mx - TILT and (x + 1 - first) / sc > mn + TILT
    ]
  elif side == 3: # west
    sc = ref.referent.vt_scale
    first = offy * sc
    side = [
      ref.referent.get(0, y)
      for y in sorted(
        range(first, first + sc),
        key=lambda i: abs((i - first + 0.5)/sc - biasNS)
      )
      if (y - first) / sc < mx - TILT and (y + 1 - first) / sc > mn + TILT
    ]
  # Decide where to enter:
  enp = None
  for px in side:
    if clear_test(px):
      enp = px
      break
  return enp

def exit_point(pos, edge, dir):
  '''
  A helper function for move_position that computes the exit point from a
  surface given the position, the particular referent to the edge moved out on,
  and the direction of movement.
  '''
  if not edge.is_edge:
    raise TypeError("Can't move out at a non-edge.")
  if not pos.get_parent():
    return None
  # Find the root reference in the parent surface:
  root = frwrap.Surface.find_root_of(pos.get_parent().fraxel)
  # Compute offsets:
  offx = edge.x // edge.surface.hz_scale
  offy = edge.y // edge.surface.vt_scale

  w = edge.surface.sc_width
  h = edge.surface.sc_height

  if root.orientation == 'n':
    off_rl = offx
    off_ud = offy
  elif root.orientation == 'e':
    off_rl = h - 1 - offy
    off_ud = offx
  elif root.orientation == 's':
    off_rl = w - 1 - offx
    off_ud = h - 1 - offy
  elif root.orientation == 'w':
    off_rl = offy
    off_ud = w - 1 - offx

  x, y = root.x + off_rl, root.y + off_ud

  return root.surface.get(x, y)


################
# Rule classes #
################

class Rule:
  '''
  A rule is a function that applies either globally or to an individual Fraxel.
  Rules may be setup rules, which are fired once at startup, world rules, which
  are fired once every step, or scan rules, which are fired once for each
  Fraxel of each surface at every step. Each rule has an 'order' attribute
  which determines when it gets executed relative to other rules, and a name
  attribute (which will be deduced from the name of the given function if it
  isn't passed to the constructor). The order may be positive or negative:
  rules are executed according to their order values, starting from zero and
  counting up to the highest-order rule, and then starting from the most
  negative-order rule and counting up to the least negative-ordered rule. For
  example, given rules 'A', 'B', 'C', and 'D' with orders 0, 1, -1, and -0.5,
  the order of execution will be 'A', 'B', 'C', 'D'.
  '''
  fn = lambda: None
  name = "<undefined>"
  order = 1

  def __init__(self, fn, name=None, order=1):
    self.fn = fn
    self.name = name or fn.__code__.co_name
    self.order = order
    self.is_local = False
    if self.fn.__code__.co_argcount:
      self.is_local = True

  def __str__(self):
    return "the {} rule".format(self.name)

  def __repr__(self):
    return "<rule '{}'>".format(self.name)

  def __call__(self, fraxel=None):
    if self.fn.__code__.co_argcount == 1:
      results = self.fn(fraxel)
    else:
      results = self.fn()
    return results

def rule(order=1, name=None):
  '''
  A decorator for turning functions into rules. Takes the desired order and
  optionally a name for the rule to be constructed. Note that this is actually
  both a decorator and a decorator producer, and as such may be invoked as
  either @rule(order, name) or just @rule.
  '''
  if isinstance(order, function):
    return Rule(order)
  # else:
  def decorator(fn):
    return Rule(fn, name=name, order=order)
  return decorator

######################
# The step procedure #
######################

def apply_rules(rules, world=None):
  '''
  Applies the given list of rules to the given world, according to their order
  values.
  '''
  if not rules:
    return
  orders = [r.order for r in rules]
  mx = max(orders)
  mn = min(orders)
  for r in sorted(
    list(rules),
    key=lambda r: r.order + (r.order < 0 and (mx - mn + 1))
  ):
    if r.is_local:
      if not world:
        raise ValueError("Tried to apply a local rule without a world.")
      for name in world:
        if name == "<name>":
          continue
        surface = world[name]
        for x in range(surface.width):
          for y in range(surface.height):
            r(surface.get(x, y))
    else:
      r()
    debug.debug(
      "rules",
      "Rule '{}' applied.".format(r.name)
    )

def step(rules, world):
  '''
  Steps time forward by one tick, applying all of the given rules to the given
  world.
  '''
  global SIGNALS, NEXT_SIGNALS, STEP

  debug.debug("step", "-"*73)
  debug.debug("step", "Starting step {}...".format(STEP))

  # Activate the next set of signals:
  SIGNALS = NEXT_SIGNALS
  NEXT_SIGNALS = set()

  # Apply the rules:
  apply_rules(rules, world)

  STEP += 1

  debug.debug("step", "...done with step {}.".format(STEP))
  debug.debug("step", "-"*73)
