#include "Python.h"
#include "structmember.h"

#include <GL/gl.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

// TODO: Store fractal scale in each surface...

/**************************
 * Constants and globals: *
 **************************/

static int WINDOW_WIDTH = -1; // in pixels
static int WINDOW_HEIGHT = -1; // in pixels

static const int VIEW_SIZE = 23; // in squares

static float SQUARE_SIZE = -1; // in pixels

static float ZOOM_FACTOR = 5;
static float ZOOM = 0; // in powers of the zoom factor
static const float MAX_ZOOM = 3;
static const float MIN_ZOOM = -1;

static const int MAX_DRAW_DEPTH = 12;

/*************************
 * Forward declarations: *
 *************************/

struct Surface_s;
typedef struct Surface_s Surface;

struct Position_s;
typedef struct Position_s Position;

/************
 * Structs: *
 ************/

typedef struct Location_s {
  int x, y;
} Location;

typedef struct Pixel_s {
  unsigned char r, g, b;
} Pixel;

typedef struct Reference_s {
  Surface *referent;
  unsigned char orientation; // NESW <-> 0123 (direction of "up")
} Reference;

typedef union PixRef_u {
  Pixel pixel;
  Reference reference;
} PixRef;

typedef struct Fraxel_s {
  PixRef ref;
  unsigned char type;
} Fraxel;

// Fraxel types:
//  0 - not a reference.
//  1 - a normal reference.
//  2 - a subordinate reference.

struct Surface_s {
  PyObject_HEAD
  unsigned char hz_scale, vt_scale; // Fraxels per scale unit.
  unsigned char sc_width, sc_height; // In terms of scale units.
  Fraxel *fraxels;
};

struct Position_s {
  PyObject_HEAD
  int x, y;
  Position *parent;
  Surface *surface;
};

/********************************
 * Data manipulation functions: *
 ********************************/

static float getViewSize(void) {
  return (VIEW_SIZE * pow(ZOOM_FACTOR, ZOOM));
}

/*
 * Looks up a fraxel by x- and y-coordinate within a surface.
 */
static Fraxel* getFraxel(Surface *within, int x, int y) {
  if (x < 0 || y < 0
   || x >= within->sc_width * within->hz_scale
   || y >= within->sc_height * within->vt_scale) {
    return NULL;
  }
  return &(within->fraxels[
    x + (int) (y * within->sc_width * within->hz_scale)
  ]);
}

/*
 * Finds the root of a secondary reference. Returns NULL for Pixels, and the
 * location of the given fraxel for primary references.
 */
static Location* findRoot(
  Surface *within,
  int x, int y,
  int ox, int oy,
  int onlywest
) {
  Surface *referent;
  Location *result, *north;
  Fraxel *frx;
  frx = getFraxel(within, x, y);
  if (frx == NULL || frx->type == 0) {
    return NULL;
  } else if (frx->type == 1) {
    referent = frx->ref.reference.referent;
    if (frx->ref.reference.orientation % 2) {
      if (x + referent->sc_height > ox && y + referent->sc_width > oy) {
        result = PyMem_New(Location, 1);
        result->x = x;
        result->y = y;
        return result;
      }
    } else {
      if (x + referent->sc_width > ox && y + referent->sc_height > oy) {
        result = PyMem_New(Location, 1);
        result->x = x;
        result->y = y;
        return result;
      }
    }
    return NULL;
  } else {
    if (onlywest) {
      return findRoot(within, x - 1, y, ox, oy, 1);
    } else {
      north = findRoot(within, x, y - 1, ox, oy, 0);
      if (north != NULL) {
        return north;
      }
      return findRoot(within, x - 1, y, ox, oy, 1);
    }
  }
}

/******************************
 * View management functions: *
 ******************************/

/*
 * Sets the perspective
 */
static void setPerspective(void) {
  float viewsize = getViewSize();
  // Computer pixels per square:
  if (WINDOW_WIDTH >= WINDOW_HEIGHT) {
    SQUARE_SIZE = WINDOW_HEIGHT / viewsize;
  } else {
    SQUARE_SIZE = WINDOW_WIDTH / viewsize;
  }

  // Initial matrix setup:
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();

  /*
   * Orthographic projection. The arguments are in order the left, right,
   * bottom, top, near and far clipping planes. Note that the top and bottom
   * clipping planes have been swapped for the sake of sanity.
   */
  glOrtho(0, viewsize, viewsize, 0, -1, 1);
}

static void setZoom(float zoom) {
  if (zoom > MAX_ZOOM) {
    ZOOM = MAX_ZOOM;
  } else if (zoom < MIN_ZOOM) {
    ZOOM = MIN_ZOOM;
  } else {
    ZOOM = zoom;
  }
  setPerspective();
}

static float getZoom(void) {
  return ZOOM;
}

/*
 * Resizes the view according to new window width and height parameters.
 */
static void resize(int w, int h) {
  // Prevent crash in the edge case:
  if (h == 0) {
    h = 1;
  }

  WINDOW_WIDTH = w;
  WINDOW_HEIGHT = h;

  // Reset the viewport:
  if (w >= h) {
    glViewport((w - h)/2, 0, h, h);
  } else {
    glViewport(0, (h - w)/2, w, w);
  }

  // Set the perspective:
  setPerspective();
}

/**********************
 * Drawing functions: *
 **********************/

static void translateInto(Reference *reference, int x, int y) {
  glTranslatef(x, y, 0);
  if (reference->orientation == 1) {
    glRotatef(90, 0, 0, 1);
    glTranslatef(0, -reference->referent->sc_height, 0);
  } else if (reference->orientation == 2) {
    glRotatef(180, 0, 0, 1);
    glTranslatef(
      -reference->referent->sc_width,
      -reference->referent->sc_height,
      0
    );
  } else if (reference->orientation == 3) {
    glRotatef(-90, 0, 0, 1);
    glTranslatef(-reference->referent->sc_width, 0, 0);
  }
  glScalef(
    1/((float) reference->referent->hz_scale),
    1/((float) reference->referent->vt_scale),
    0
  );
}

static void translateOutOf(Reference *reference, int x, int y) {
  glScalef(
    reference->referent->hz_scale,
    reference->referent->vt_scale,
    0
  );
  if (reference->orientation == 1) {
    glTranslatef(0, reference->referent->sc_height, 0);
    glRotatef(-90, 0, 0, 1);
  } else if (reference->orientation == 2) {
    glTranslatef(
      reference->referent->sc_width,
      reference->referent->sc_height,
      0
    );
    glRotatef(180, 0, 0, 1);
  } else if (reference->orientation == 3) {
    glTranslatef(reference->referent->sc_width, 0, 0);
    glRotatef(90, 0, 0, 1);
  }
  glTranslatef(-x, -y, 0);
}

/*
 * Secondary drawing function: renders a surface, and all subsurfaces until
 * either the maximum recursion depth is exceeded, or rendering further
 * surfaces would occur at sub-pixel levels.
 */
static void drawSurface(
  Surface *surface,
  float hz_scale,
  float vt_scale,
  int rdepth
) {
  float squarewidth = SQUARE_SIZE * hz_scale;
  float squareheight = SQUARE_SIZE * vt_scale;
  int x, y;
  Fraxel *frx;
  if (rdepth > MAX_DRAW_DEPTH) {
    return;
  }
  // TODO: Construct display lists for each Surface instead?
  for (x = 0; x < surface->sc_width * surface->hz_scale; ++x) {
    for (y = 0; y < surface->sc_height * surface->vt_scale; ++y) {
      frx = getFraxel(surface, x, y);
      // TODO: For large embeddings, this could be problematic...
      if (frx->type == 0) {
        glColor3ub(
          frx->ref.pixel.r,
          frx->ref.pixel.g,
          frx->ref.pixel.b
        );
        glBegin( GL_QUADS );
        glVertex3f(x, y, 0);
        glVertex3f(x + 1, y, 0);
        glVertex3f(x + 1, y + 1, 0);
        glVertex3f(x, y + 1, 0);
        glEnd();
      } else if (squarewidth * frx->ref.reference.referent->sc_width < 1
              || squareheight * frx->ref.reference.referent->sc_height < 1) {
        glColor3ub(128, 128, 128);
        glBegin( GL_QUADS );
        glVertex3f(x, y, 0);
        glVertex3f(x + 1, y, 0);
        glVertex3f(x + 1, y + 1, 0);
        glVertex3f(x, y + 1, 0);
        glEnd();
      } else if (frx->type == 1) {
        glPushMatrix();
        translateInto(&(frx->ref.reference), x, y);
        drawSurface(
          frx->ref.reference.referent,
          hz_scale * 1/((float) frx->ref.reference.referent->hz_scale),
          vt_scale * 1/((float) frx->ref.reference.referent->vt_scale),
          rdepth + 1
        );
        glPopMatrix();
      }
    }
  }
}

/*
 * Primary drawing function: renders a Position, plus any sub-surfaces until
 * further reference rendering would be at sub-pixel values, and any parent
 * surfaces (defined by parent positions) until further rendering would
 * necessarily be outside the screen. To do this, the appropriate number of
 * parent positions should exist, as constrained by the fractal scale, the
 * VIEW_SIZE, and the ZOOM. Otherwise, the program will only draw as many
 * parents as it can find. A previous rotation and scale may be given along
 * with an age between 0 and 1 in which case the perspective will be a mixture
 * of the old and new perspectives weighted according to the age (0 = just the
 * old perspective, 1 = just the new perspective).
 */
static void render(
  Position *target,
  float prev_rot,
  float prev_hz_scale,
  float prev_vt_scale,
  float age
) {
  float hz_scale, vt_scale;
  float center;
  float viewsize = getViewSize();
  float corr_rot, corr_hz_scale, corr_vt_scale;
  Position *tmp, *outer;
  Location *root;

  // Clear the color buffer:
  glClear( GL_COLOR_BUFFER_BIT );

  // Set up a fresh model view:
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Draw in the background for the active part of the window:
  glBegin( GL_QUADS );

  // The background is actually a light grey rather than black (but it rarely
  // gets shown)...
  glColor3ub(212, 212, 212);

  glVertex3f(0, 0, 0);
  glVertex3f(viewsize, 0, 0);
  glVertex3f(viewsize, viewsize, 0);
  glVertex3f(0, viewsize, 0);

  glEnd();

  // Compute the center:
  center = viewsize / 2.0;

  // Translate so that (0, 0) is centered:
  glTranslatef(
    center - 0.5,
    center - 0.5,
    0
  );

  // Correct for the previous position:
  // TODO: These rules (or their drivers) are *almost* but not quite correct.
  if (age < 1) {
    corr_rot = prev_rot * (1.0 - age);
    if (corr_rot > 180) {
      corr_rot -= 180;
    }
    corr_hz_scale = pow(prev_hz_scale, 1.0 - age);
    corr_vt_scale = pow(prev_vt_scale, 1.0 - age);
    glScalef(corr_hz_scale, corr_vt_scale, 0);
    glRotatef(corr_rot, 0, 0, 1);
  }

  // Draw the world, centered on the given position, starting with the
  // theoretical outermost surface (as constrained by the VIEW_SIZE,
  // fractal scales, and ZOOM).
  hz_scale = 1.0;
  vt_scale = 1.0;
  tmp = target;
  // Translate to the origin of the initial surface:
  glTranslatef(-tmp->x, -tmp->y, 0);
  while ( viewsize / hz_scale > 1 || viewsize / vt_scale > 1) {
    outer = tmp->parent;
    if (outer == NULL) {
      outer = tmp;
      break;
    }
    hz_scale *= tmp->surface->hz_scale;
    vt_scale *= tmp->surface->vt_scale;
    root = findRoot(
      outer->surface,
      outer->x,
      outer->y,
      outer->x,
      outer->y,
      0
    );
    translateOutOf(
      &(getFraxel(outer->surface, outer->x, outer->y)->ref.reference),
      root->x,
      root->y
    );
    PyMem_Free(root);
    tmp = outer;
  }
  // outer is now the outermost Surface to draw, and the model view matrix is
  // positioned accordingly. Simply draw the Surface and all subsurfaces until
  // either an appropriate resolution is achieved or we run out of allowed
  // depth.
  drawSurface(outer->surface, hz_scale, vt_scale, 1);
}

/*****************************
 * Python object structures: *
 *****************************/

// Forward declarations for the Python Type structs:

static PyTypeObject PositionType;
static PyTypeObject SurfaceType;

/**************************
 * Python type functions: *
 **************************/

static void Position_dealloc(Position *self) {
  Py_CLEAR(self->parent);
  Py_CLEAR(self->surface);
  Py_TYPE(self)->tp_free((PyObject*)self);
}

static PyObject* Position_new(
  PyTypeObject *type,
  PyObject *args,
  PyObject *kwds
) {
  Position *self;

  self = (Position*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->x = -1;
    self->y = -1;
    self->parent = NULL;
    self->surface = NULL;
  }

  return (PyObject*) self;
}

static int Position_init(
  Position *self,
  PyObject *args,
  PyObject *kwds
) {
  static char *kwlist[] = {"x", "y", "surface", "parent", NULL};
  self->parent = NULL;

  if (!PyArg_ParseTupleAndKeywords(
    args,
    kwds,
    "iiO!|O!",
    kwlist,
    &(self->x),
    &(self->y),
    &SurfaceType,
    &(self->surface),
    &PositionType,
    &(self->parent)
  )) {
    return -1;
  }
  Py_INCREF(self->surface);
  Py_XINCREF(self->parent);
  return 0;
}

// Returns a new reference.
static PyObject* Position_getsurface(Position *self, void *closure)
{
  if (self->surface == NULL) {
    Py_RETURN_NONE;
  }
  Py_INCREF(self->surface);
  return (PyObject*) (self->surface);
}

static int Position_setsurface(
  Position *self,
  PyObject *value,
  void *closure
) {
  if (value == NULL) {
    PyErr_SetString(PyExc_TypeError, "Cannot delete surface attribute.");
    return -1;
  }

  if (! PyObject_TypeCheck(value, &SurfaceType)) {
    PyErr_SetString(
      PyExc_TypeError,
      "The surface attribute must hold a Surface."
    );
    return -1;
  }

  Py_INCREF(value);
  Py_CLEAR(self->surface);
  self->surface = (Surface*) value;

  return 0;
}

// Returns a new reference.
static PyObject* Position_getparent(Position *self, void *closure)
{
  if (self->parent == NULL) {
    Py_RETURN_NONE;
  }
  Py_INCREF(self->parent);
  return (PyObject*) (self->parent);
}

static int Position_setparent(
  Position *self,
  PyObject *value,
  void *closure
) {
  if (value == Py_None) {
    value = NULL;
  }
  if (value != NULL && !PyObject_TypeCheck(value, &PositionType)) {
    PyErr_SetString(
      PyExc_TypeError,
      "The parent attribute must hold a Surface."
    );
    return -1;
  }

  Py_XINCREF(value);
  Py_CLEAR(self->parent);
  self->parent = (Position*) value;

  return 0;
}

static void Surface_dealloc(Surface *self) {
  if (self->fraxels != NULL) {
    PyMem_Free((void*) self->fraxels);
  }
  Py_TYPE(self)->tp_free((PyObject*)self);
}

static PyObject* Surface_new(
  PyTypeObject *type,
  PyObject *args,
  PyObject *kwds
) {
  Surface *self;

  self = (Surface*) type->tp_alloc(type, 0);
  if (self != NULL) {
    self->sc_height = -1;
    self->sc_width = -1;
    self->fraxels = NULL;
  }

  return (PyObject*) self;
}

static int Surface_init(
  Surface *self,
  PyObject *args,
  PyObject *kwds
) {
  int i, j, w, h;
  static char *kwlist[] = {
    "hz_scale",
    "vt_scale",
    "sc_width",
    "sc_height",
    NULL
  };
  self->hz_scale = 5;
  self->vt_scale = 5;
  self->sc_width = 1;
  self->sc_height = 1;

  if (!PyArg_ParseTupleAndKeywords(
    args,
    kwds,
    "|iiii",
    kwlist,
    &(self->hz_scale),
    &(self->vt_scale),
    &(self->sc_width),
    &(self->sc_height)
  )) {
    return -1;
  }

  w = (int) (self->sc_width * self->hz_scale);
  h = (int) (self->sc_height * self->vt_scale);

  self->fraxels = PyMem_New(Fraxel, w * h);

  if (self->fraxels == NULL) {
    PyErr_NoMemory();
    return -1;
  }

  for (i = 0; i < w; ++i) {
    for (j = 0; j < h; ++j) {
      self->fraxels[i + j*w].type = 0;
      self->fraxels[i + j*w].ref.pixel.r = 0;
      self->fraxels[i + j*w].ref.pixel.g = 0;
      self->fraxels[i + j*w].ref.pixel.b = 0;
    }
  }

  return 0;
}

// Returns a new reference.
static PyObject* Surface_get(Surface *self, PyObject *args, PyObject *kwds) {
  int x, y, w, h;
  Fraxel *frx;
  static char *kwlist[] = {"x", "y", NULL};

  if (!PyArg_ParseTupleAndKeywords(
    args, kwds,
    "ii",
    kwlist,
    &x, &y
  )) {
    return NULL;
  }

  w = (int) (self->sc_width * self->hz_scale);
  h = (int) (self->sc_height * self->vt_scale);

  if (x < 0) { x += w; }
  if (y < 0) { y += h; }

  if (x < 0 || x >= w || y < 0 || y >= h) {
    Py_RETURN_NONE;
  }

  frx = &(self->fraxels[x + y*w]);
  if (frx->type == 0) {
    return Py_BuildValue(
      "(iii)",
      (int) frx->ref.pixel.r, // TODO: Do these casts do the right thing?
      (int) frx->ref.pixel.g,
      (int) frx->ref.pixel.b
    );
  } else {
    return Py_BuildValue(
      "(Oi)",
      (PyObject*) (frx->ref.reference.referent),
      frx->ref.reference.orientation
    );
  }
}

static PyObject* Surface_set(Surface *self, PyObject *args, PyObject *kwds) {
  int x, y, xi, yi, w, h, orientation, right, top, oldright, oldtop;
  unsigned char r = 0, g = 0, b = 0;
  long lr, lg, lb;
  PyObject *value, *pyr, *pyg, *pyb, *arguments, *result;
  Surface *old;
  Fraxel *frx;
  static char *kwlist[] = {"x", "y", "value", "orientation", NULL};

  orientation = 0;

  if (!PyArg_ParseTupleAndKeywords(
    args, kwds,
    "iiO|i",
    kwlist,
    &x, &y, &value, &orientation
  )) {
    return NULL;
  }

  if (orientation < 0 || orientation > 3) {
    PyErr_Format(
      PyExc_ValueError,
      "Orientation %d is not valid (valid values are 0, 1, 2, 3).",
      orientation
    );
    return NULL;
  }

  right = x;
  top = y;

  w = (int) (self->sc_width * self->hz_scale);
  h = (int) (self->sc_height * self->vt_scale);

  if (x < 0) { x += w; }
  if (y < 0) { y += h; }

  if (x < 0 || x >= w || y < 0 || y >= h) {
    PyErr_Format(
      PyExc_IndexError,
      "Fraxel at (%d, %d) is out-of-bounds (surface is %dx%d).",
      x, y, w, h
    );
    return NULL;
  }

  // Type checking of the value argument (and r, g, b unpacking):
  // First, set up a generic error string (cleared later if checks succeed):
  PyErr_SetString(
    PyExc_TypeError,
    "Value parameter must be either a Surface or a sequence of three integers."
  );
  if (value == NULL) {
    return NULL;
    // Value is NULL (shouldn't happen).
  } else if (PySequence_Check(value)) {
    if (PySequence_Size(value) != 3) {
      // Sequence has a length other than 3.
      return NULL;
    } else {
      pyr = PySequence_GetItem(value, 0);
      pyg = PySequence_GetItem(value, 1);
      pyb = PySequence_GetItem(value, 2);
      if (PyLong_Check(pyr) && PyLong_Check(pyg) && PyLong_Check(pyb)) {
        lr = PyLong_AsLong(pyr);
        lg = PyLong_AsLong(pyg);
        lb = PyLong_AsLong(pyb);
        if (0 <= lr && lr < 256 && 0 <= lg && lg < 256 && 0 <= lb && lb < 256) {
          r = (unsigned char) lr;
          g = (unsigned char) lg;
          b = (unsigned char) lb;
        } else {
          PyErr_SetString(
            PyExc_ValueError,
            "Color values must be in [0-255]."
          );
          // Bad color value.
          return NULL;
        }
      } else {
        // Sequence value isn't an integer.
        return NULL;
      }
    }
  } else if (PyObject_TypeCheck(value, &SurfaceType)) {
    if (orientation % 2) {
      right = x + ((Surface*) value)->sc_height - 1;
      top = y + ((Surface*) value)->sc_width - 1;
    } else {
      right = x + ((Surface*) value)->sc_width - 1;
      top = y + ((Surface*) value)->sc_height - 1;
    }
    if (right >= w || top >= h) {
      PyErr_Format(
        PyExc_ValueError,
        "Reference is too large to be placed at (%d, %d)",
        x, y
      );
      // Reference is too large.
      return NULL;
    }
    // Clear out space for the reference (we'll hit an error here if it would
    // overlap another reference awkwardly).
    for (xi = x; xi <= right; ++xi) {
      for (yi = y; yi <= top; ++yi) {
        arguments = Py_BuildValue("(ii(iii))", xi, yi, 0, 0, 0);
        Py_INCREF(arguments);
        result = Surface_set(self, arguments, Py_None);
        Py_CLEAR(result);
        Py_CLEAR(arguments);
      }
    }
  } else {
    // Argument is neither a sequence nor a Surface.
    return NULL;
  }
  PyErr_Clear();

  // Type-independent fraxel lookup:
  frx = &(self->fraxels[x + y*w]);
  if (frx->type != 0 && frx->type != 1) {
    PyErr_SetString(
      PyExc_ValueError,
      "Only pixel and primary reference fraxels may be changed."
    );
    return NULL;
  }

  // If we're overwriting a reference, clear it out first:
  if (frx->type == 1) {
    old = frx->ref.reference.referent;
    if (frx->ref.reference.orientation % 2) {
      oldright = x + old->sc_height - 1;
      oldtop = y + old->sc_width - 1;
    } else {
      oldright = x + old->sc_width - 1;
      oldtop = y + old->sc_height - 1;
    }
    for (xi = x; xi <= oldright; ++xi) {
      for (yi = y; yi <= oldtop; ++yi) {
        if (xi != x || yi != y) {
          self->fraxels[xi + yi*w].type = 0;
          self->fraxels[xi + yi*w].ref.pixel.r = 0;
          self->fraxels[xi + yi*w].ref.pixel.g = 0;
          self->fraxels[xi + yi*w].ref.pixel.b = 0;
        }
      }
    }
    Py_CLEAR(frx->ref.reference.referent);
  }

  if (PyObject_TypeCheck(value, &SurfaceType)) {
    Py_INCREF(value);
    frx->type = 1;
    frx->ref.reference.orientation = orientation;
    frx->ref.reference.referent = (Surface*) value;
    for (xi = x; xi <= right; ++xi) {
      for (yi = y; yi <= top; ++yi) {
        if (xi != x || yi != y) {
          self->fraxels[xi + yi*w].type = 2;
          self->fraxels[xi + yi*w].ref.reference.orientation = orientation;
          self->fraxels[xi + yi*w].ref.reference.referent = (Surface*) value;
        }
      }
    }
  } else {
    frx->type = 0;
    frx->ref.pixel.r = r;
    frx->ref.pixel.g = g;
    frx->ref.pixel.b = b;
  }

  Py_RETURN_NONE;
}

// Returns a new reference.
static PyObject* Surface_is_secondary(
  Surface *self,
  PyObject *args,
  PyObject *kwds
) {
  int x, y, w, h;
  Fraxel *frx;
  static char *kwlist[] = {"x", "y", NULL};

  if (!PyArg_ParseTupleAndKeywords(
    args, kwds,
    "ii",
    kwlist,
    &x, &y
  )) {
    return NULL;
  }

  w = (int) (self->sc_width * self->hz_scale);
  h = (int) (self->sc_height * self->vt_scale);

  if (x < 0) { x += w; }
  if (y < 0) { y += h; }

  if (x < 0 || x >= w || y < 0 || y >= h) {
    Py_RETURN_FALSE;
  }

  frx = &(self->fraxels[x + y*w]);
  if (frx->type == 2) {
    Py_RETURN_TRUE;
  } else {
    Py_RETURN_FALSE;
  }
}

// Returns a new reference.
static PyObject* Surface_find_root(
  Surface *self,
  PyObject *args,
  PyObject *kwds
) {
  int x, y;
  Location *result;
  PyObject *ret;
  static char *kwlist[] = {"x", "y", NULL};

  if (!PyArg_ParseTupleAndKeywords(
    args, kwds,
    "ii",
    kwlist,
    &x, &y
  )) {
    return NULL;
  }

  result = findRoot(self, x, y, x, y, 0);
  
  if (result == NULL) {
    Py_RETURN_NONE;
  }

  ret = Py_BuildValue("(ii)", result->x, result->y);

  PyMem_Free(result);

  return ret;
}


/**********************
 * Python type setup: *
 **********************/

static PyMemberDef Position_members[] = {
  { "x", T_INT, offsetof(Position, x), 0, "The Position's x-coordinate." },
  { "y", T_INT, offsetof(Position, y), 0, "The Position's y-coordinate." },
  {NULL}
};

static PyMemberDef Surface_members[] = {
  { "hz_scale", T_UBYTE, offsetof(Surface, hz_scale), READONLY,
    "The Surface horizontal scale (fraxels per scale unit)." },
  { "vt_scale", T_UBYTE, offsetof(Surface, vt_scale), READONLY,
    "The Surface vertical scale (fraxels per scale unit)." },
  { "sc_width", T_UBYTE, offsetof(Surface, sc_width), READONLY,
    "The Surface width (in scale units)." },
  { "sc_height", T_UBYTE, offsetof(Surface, sc_height), READONLY,
    "The Surface height (in scale units)." },
  {NULL}
};

static PyGetSetDef Position_getseters[] = {
  {
    "surface",
    (getter)Position_getsurface,
    (setter)Position_setsurface,
    "The Surface within which this Position is located.",
    NULL
  },
  {
    "parent",
    (getter)Position_getparent,
    (setter)Position_setparent,
    "The parent of this Position.",
    NULL
  },
  {NULL}
};

static PyMethodDef Surface_methods[] = {
  { "get", (PyCFunction)Surface_get, METH_VARARGS | METH_KEYWORDS,
    "Arguments: x, y\n"
    "Returns the fraxel at the given x and y coordinates. Handles negative "
    "indices like a list does. Returns a Surface, orientation pair (for "
    "primary and secondary references), a 3-tuple of integers in [0-255] (for "
    "colors), or None (for indices outside of the Surface)." },
  { "set", (PyCFunction)Surface_set, METH_VARARGS | METH_KEYWORDS,
    "Arguments: x, y, value\n"
    "Sets the fraxel at the given coordinates. Value should be either a "
    "Surface or a 3-tuple of integers in [0-255] (R, G, and B components of a "
    "24-bit color. Note that if a Surface is given, it may change more than "
    "just the given fraxel, since scale size is maintained. The given "
    "coordinates are the minimum x and y values for the new reference. Also, "
    "it's possible for the new reference to overlap another reference or cross "
    "the edge of the Surface. In these cases, an exception is raised, even if "
    "the given x- and y-coordinates are valid as a singe fraxel. Also note "
    "that no operation is allowed to change a secondary reference fraxel. "
    "Instead, the primary reference must first be changed, which will get rid "
    "of any associated secondary references (setting them to (0, 0, 0))." },
  { "is_secondary", (PyCFunction)Surface_is_secondary,
    METH_VARARGS | METH_KEYWORDS,
    "Arguments: x, y\n"
    "If the given fraxel is a secondary reference, returns True. If it is "
    "anything else (a primary reference, a pixel, or out-of-bounds), it "
    "returns False." },
  { "find_root", (PyCFunction)Surface_find_root,
    METH_VARARGS | METH_KEYWORDS,
    "Arguments: x, y\n"
    "Returns a pair of coordinates that indicate the location within this "
    "Surface of the root of the reference at the given coordinates. Returns "
    "None if the Fraxel at the given coordinates isn't a reference." },
  { NULL }
};

static PyTypeObject PositionType = {
    PyVarObject_HEAD_INIT(NULL, 0)
    "fractal.Position",               /* tp_name */
    sizeof(Position),                 /* tp_basicsize */
    0,                                /* tp_itemsize */
    (destructor)Position_dealloc,     /* tp_dealloc */
    0,                                /* tp_print */
    0,                                /* tp_getattr */
    0,                                /* tp_setattr */
    0,                                /* tp_reserved */
    0,                                /* tp_repr */
    0,                                /* tp_as_number */
    0,                                /* tp_as_sequence */
    0,                                /* tp_as_mapping */
    0,                                /* tp_hash  */
    0,                                /* tp_call */
    0,                                /* tp_str */
    0,                                /* tp_getattro */
    0,                                /* tp_setattro */
    0,                                /* tp_as_buffer */
    Py_TPFLAGS_DEFAULT |
        Py_TPFLAGS_BASETYPE,          /* tp_flags */
    "A position in fractal space. Takes x, y, surface, and parent arguments "
    "for construction (the last of which is optional). The surface argument "
    "must be a Surface, and the parent argument, if given, must be a "
    "Position.",                      /* tp_doc */
    0,		                      /* tp_traverse */
    0,		                      /* tp_clear */
    0,		                      /* tp_richcompare */
    0,		                      /* tp_weaklistoffset */
    0,		                      /* tp_iter */
    0,		                      /* tp_iternext */
    0,                                /* tp_methods */
    Position_members,                 /* tp_members */
    Position_getseters,               /* tp_getset */
    0,                                /* tp_base */
    0,                                /* tp_dict */
    0,                                /* tp_descr_get */
    0,                                /* tp_descr_set */
    0,                                /* tp_dictoffset */
    (initproc)Position_init,          /* tp_init */
    0,                                /* tp_alloc */
    (newfunc)Position_new             /* tp_new */
    // TODO: Implement cyclic garbage collection!
};

static PyTypeObject SurfaceType = {
    PyVarObject_HEAD_INIT(NULL, 0)
    "fractal.Surface",                /* tp_name */
    sizeof(Surface),                  /* tp_basicsize */
    0,                                /* tp_itemsize */
    (destructor)Surface_dealloc,      /* tp_dealloc */
    0,                                /* tp_print */
    0,                                /* tp_getattr */
    0,                                /* tp_setattr */
    0,                                /* tp_reserved */
    0,                                /* tp_repr */
    0,                                /* tp_as_number */
    0,                                /* tp_as_sequence */
    0,                                /* tp_as_mapping */
    0,                                /* tp_hash  */
    0,                                /* tp_call */
    0,                                /* tp_str */
    0,                                /* tp_getattro */
    0,                                /* tp_setattro */
    0,                                /* tp_as_buffer */
    Py_TPFLAGS_DEFAULT |
        Py_TPFLAGS_BASETYPE,          /* tp_flags */
    "A fractal surface (composed of fraxels). It takes hz_scale, vt_scale, "
    "sc_width and sc_height arguments (each optional) in the constructor, and "
    "constructs a surface that is the requested width and height, with the "
    "requested fractal scales. So for example, if the call is: "
    "Surface(3, 4, 3, 5), the resulting surface will be 9 fraxels wide and 20 "
    "fraxels tall, and will take up 3x5 fraxels when referred to. When "
    "referenced, a surface always takes up exactly sc_width by sc_height "
    "fraxels (that's what the sc_ means). Note that the all four arguments "
    "should be in [0-255]. When initialized, each fraxel in the surface is set "
    "to (0, 0, 0).",                  /* tp_doc */
    0,		                      /* tp_traverse */
    0,		                      /* tp_clear */
    0,		                      /* tp_richcompare */
    0,		                      /* tp_weaklistoffset */
    0,		                      /* tp_iter */
    0,		                      /* tp_iternext */
    Surface_methods,                  /* tp_methods */
    Surface_members,                  /* tp_members */
    0,                                /* tp_getset */
    0,                                /* tp_base */
    0,                                /* tp_dict */
    0,                                /* tp_descr_get */
    0,                                /* tp_descr_set */
    0,                                /* tp_dictoffset */
    (initproc)Surface_init,           /* tp_init */
    0,                                /* tp_alloc */
    (newfunc)Surface_new              /* tp_new */
    // TODO: Implement cyclic garbage collection!
};

/****************************
 * Python module functions: *
 ****************************/

static PyObject* py_setZoom(
  PyObject *self,
  PyObject *args,
  PyObject *kwds
) {
  float zoom;
  static char *kwlist[] = {"zoom", NULL};

  if (!PyArg_ParseTupleAndKeywords(
    args, kwds,
    "f",
    kwlist,
    &zoom
  )) {
    return NULL;
  }

  setZoom(zoom);

  Py_RETURN_NONE;
}

static PyObject* py_getZoom(PyObject *self) {
  return Py_BuildValue("f", getZoom());
}

static PyObject* py_resize(
  PyObject *self,
  PyObject *args,
  PyObject *kwds
) {
  int w, h;
  static char *kwlist[] = {"width", "height", NULL};

  if (!PyArg_ParseTupleAndKeywords(
    args, kwds,
    "ii",
    kwlist,
    &w, &h
  )) {
    return NULL;
  }

  resize(w, h);

  Py_RETURN_NONE;
}

static PyObject* py_render(
  PyObject *self,
  PyObject *args,
  PyObject *kwds
) {
  PyObject* position;
  float last_angle = 0, last_hz_scale = 1, last_vt_scale = 1, age=1;

  static char *kwlist[] = {
    "position",
    "last_angle",
    "last_hz_scale",
    "last_vt_scale",
    "age",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(
    args, kwds,
    "O!|ffff",
    kwlist,
    &PositionType,
    &position,
    &last_angle,
    &last_hz_scale,
    &last_vt_scale,
    &age
  )) {
    return NULL;
  }

  render((Position*) position, last_angle, last_hz_scale, last_vt_scale, age);

  Py_RETURN_NONE;
}

static PyObject* py_getViewSize(PyObject *self) {
  return Py_BuildValue("f", getViewSize());
}

static PyMethodDef FractalMethods[] = {
  { "get_zoom", (PyCFunction)py_getZoom, METH_NOARGS,
    "No arguments.\n"
    "Gets the viewing zoom level. Returns a floating point number. Zoom levels "
    "are powers of the fractal scale, with 0 being no zoom, and negative "
    "numbers as increasing magnification."},
  { "set_zoom", (PyCFunction)py_setZoom, METH_VARARGS | METH_KEYWORDS,
    "Arguments: zoom\n"
    "Sets the viewing zoom level. Takes a single floating point number. Zoom "
    "levels are powers of the fractal scale, with 0 being no zoom, and "
    "negative numbers as increasing magnification. Zoom values are silently "
    "clamped to a predefined range; use get_zoom to check what the actual "
    "current zoom value is after setting it." },
  { "resize", (PyCFunction)py_resize, METH_VARARGS | METH_KEYWORDS,
    "Arguments: width, height\n"
    "An OpenGL callback for resizing the window. This should be called at the "
    "start of the program (after OpenGL context is set up), and then used as "
    "(at least part of) the OpenGL reshape callback." },
  { "render", (PyCFunction)py_render, METH_VARARGS | METH_KEYWORDS,
    "Arguments: position, [last_angle, last_hz_scale, last_vt_scale, age]\n"
    "Calls OpenGL functions to render a fractal world, centered on the given "
    "position. If the optional arguments are given, the age (between 0 (full "
    "blending) and 1 (no blending)) is used to blend the perspective for the "
    "given position. Note that this only draws as many outer surfaces as the "
    "given position has parents (up to enough to fill the screen), so ensuring "
    "that enough parents exist before calling this function is a good idea. "
    "Also note that this doesn't call glutSwapBuffers: it just performs OpenGL "
    "drawing commands. This allows other functions to be called afterwards "
    "to draw more stuff, but places the responsibility on the caller for "
    "ultimately swapping buffers." },
  { "get_view_size", (PyCFunction)py_getViewSize, METH_NOARGS,
    "No arguments.\n"
    "Returns the width of the viewport in squares (at the default scale. This "
    "should be used to ensure that positions have enough parents before calls "
    "to render." },
  { NULL }
};

/************************
 * Python module setup: *
 ************************/

static PyModuleDef fractalmodule = {
  PyModuleDef_HEAD_INIT,
  "fractal",
  "Module providing fractal data types (Position and Surface) as well as "
  "drawing routines that call OpenGL code for rendering. Doesn't include code "
  "to set up an OpenGL context.",
  -1,
  FractalMethods,
};

PyMODINIT_FUNC PyInit_fractal(void) {
  PyObject *module;
  if (PyType_Ready(&PositionType) < 0) { return NULL; }
  if (PyType_Ready(&SurfaceType) < 0) { return NULL; }

  module = PyModule_Create(&fractalmodule);

  if (module == NULL) { return NULL; }

  Py_INCREF(&PositionType);
  PyModule_AddObject(module, "Position", (PyObject*) &PositionType);
  Py_INCREF(&SurfaceType);
  PyModule_AddObject(module, "Surface", (PyObject*) &SurfaceType);

  return module;
}
