'''
frwrap.py
This module wraps the fractal module, providing extended Position and Surface
classes with extra functionality. Among other things, it provides a way to load
a collection of Surfaces from a directory as a "world", and it handles colors
as hue, saturation, and value triples with each value in [0,1].
'''

import colorsys, re

from .lib import fractal

from . import colors

from . import debug

# TODO: Round these off?
FORBIDDEN_CHARACTERS = {
  '~': "_tilde_",
  '!': "_exclaim_",
  '@': "_at_",
  '#': "_octothorpe_",
  '$': "_dollars_",
  '%': "_percent_",
  '^': "_caret_",
  '&': "_ampersand_",
  '*': "_asterisk_",
  '(': "_rparen_",
  ')': "_lparen_",
  '=': "_equals_",
  '[': "_rbrace_",
  ']': "_lbrace_",
  '{': "_rbracket_",
  '}': "_lbracket_",
  ';': "_semicolon_",
  ':': "_colon_",
  "'": "_quote_",
  '"': "_double-quote_",
  ',': "_comma_",
  '.': "_period_",
  '/': "_slash_",
  '?': "_question_",
  '\\': "_backslash_",
  '|': "_bar_",
  ' ': "_space_",
  '\n': "_newline_",
  '\t': "_tab_",
}
# '-', '_', '<', '>', and '+' are allowed

def check_name(name):
  '''
  Raises a ValueError if the given name isn't a good one.
  '''
  for c in FORBIDDEN_CHARACTERS:
    if c in name:
      raise ValueError("The '{}' character isn't allowed in names.".format(c))

def no_loader(filename):
  raise ValueError("Can't use file references: no loader supplied.")

class Fraxel:
  '''
  A Fraxel is one of a Pixel, a Reference, or an Edge. Fraxels know which
  Surface they are part of and their coordinates within that Surface. They
  have properties named for the cardinal and ordinal directions that fetch
  their neighbors. They also have properties to get groups of neighbors:
  'adjacent' returns a list of the north, east, south, and west neighbors in
  that order, 'neighbors' uses the same order (clockwise from north) and
  returns a list of all nine neighbors. 'corners' stars with northeast and
  lists just the Fraxels that touch on the diagonals. Finally, they have a
  couple of convenience functions (like a function to replace themselves).
  '''
  is_edge = False
  is_pixel = False
  is_reference = False
  def __init__(self, surface, x, y):
    self.surface = surface
    self.x = x
    self.y = y

  def __str__(self):
    name = self.__class__.__name__
    if name[0].lower() in 'aeiou':
      name = "an " + name
    else:
      name = "a " + name
    return "{} at ({}, {}) on '{}'".format(
      name,
      self.x,
      self.y,
      self.surface.name
    )

  def __repr__(self):
    return '<' + str(self) + '>'

  def __hash__(self):
    return hash(self.surface) + hash(self.x) | hash(self.y)

  def __eq__(self, other):
    if type(other) != Fraxel:
      return False
    if self.surface == other.surface\
    and self.x == other.x\
    and self.y == other.y:
      return True
    return False

  def __ne__(self, other):
    return not self == other

  @property
  def position(self):
    '''
    Returns an orphan Position pointing to this Fraxel.
    '''
    return Position(self.x, self.y, self.surface)

  def fresh(self):
    '''
    Re-reads the Fraxel data from the surface that it's on, returning a new
    Fraxel from the same location as this one.
    '''
    return self.surface.get(self.x, self.y)

  def become(self, fraxel):
    '''
    Replaces this Fraxel with the given Fraxel. Doesn't change the values of
    this Fraxel, but does make it invalid, since it's no longer an accurate
    description of the contents of its location.
    '''
    self.surface.set(self.x, self.y, fraxel)

  def update(self):
    '''
    Pushes any changes to this Fraxel onto the Surface that it belongs to.
    '''
    self.surface.set(self.x, self.y, self)

  @property
  def north(self): return self.surface.get(self.x, self.y - 1)

  @property
  def east(self): return self.surface.get(self.x + 1, self.y)

  @property
  def south(self): return self.surface.get(self.x, self.y + 1)

  @property
  def west(self): return self.surface.get(self.x - 1, self.y)

  @property
  def northeast(self):
    return self.surface.get(self.x + 1, self.y - 1)
  ne = northeast

  @property
  def southeast(self):
    return self.surface.get(self.x + 1, self.y + 1)
  se = southeast

  @property
  def southwest(self):
    return self.surface.get(self.x - 1, self.y + 1)
  sw = southwest

  @property
  def northwest(self):
    return self.surface.get(self.x - 1, self.y - 1)
  nw = northwest

  @property
  def adjacent(self): return [self.north, self.east, self.south, self.west]

  @property
  def corners(self): return [self.ne, self.se, self.sw, self.nw]

  @property
  def neighbors(self):
    return [self.n, self.ne, self.e, self.se, self.s, self.sw, self.w, self.nw]

class Edge(Fraxel):
  '''
  An Edge represents any Fraxel that is off the edge of a surface.
  '''
  is_edge = True

  def __str__(self):
    if self.surface:
      return "an Edge at ({}, {}) on '{}'".format(
        self.x,
        self.y,
        self.surface.name
      )
    else:
      return "an Edge"

  def __hash__(self):
    return super.__hash__() - 73

  def __eq__(self, other):
    if type(other) != Edge:
      return False
    if self.surface == other.surface\
    and self.x == other.x\
    and self.y == other.y:
      return True
    return False

  def __ne__(self, other):
    return not self == other

class Pixel(Fraxel, colors.Color):
  '''
  A Pixel holds color values, accessible as both normalized (floats in [0-1])
  and integer (integers according to the transformations in the colors module)
  values. See the Color class in the colors module for more detail.
  '''
  is_pixel = True
  def __init__(
    self,
    h=None, s=None, v=None,
    H=None, S=None, V=None,
    hue=None,
    surface=None,
    x=-1, y=-1
  ):
    Fraxel.__init__(self, surface, x, y)
    colors.Color.__init__(self, h, s, v, H, S, V, hue)

  def __hash__(self):
    return Fraxel.__hash__(self) - hash(self.h) | hash(self.s) | hash(self.v)

  def __eq__(self, other):
    if type(other) != Pixel:
      return False
    if self.surface == other.surface\
    and self.x == other.x\
    and self.y == other.y\
    and self.h == other.h\
    and self.s == other.s\
    and self.v == other.v:
      return True
    return False

  def __ne__(self, other):
    return not self == other

  @staticmethod
  def colored(color):
    '''
    Builds a pixel from the given Color.
    '''
    if not isinstance(color, colors.Color):
      raise ValueError("Can't build a Pixel from a non-Color.")
    return Pixel(color.h, color.s, color.v)

  def __str__(self):
    if self.surface:
      return "a Pixel ({}, {}, {}) at ({}, {}) on '{}'".format(
        self.h, self.s, self.v,
        self.x, self.y, self.surface.name,
      )
    else:
      return "a Pixel ({}, {}, {})".format(
        self.h, self.s, self.v,
      )

class Reference(Fraxel):
  '''
  A reference holds a Surface that it refers to and an orientation in which
  that surface appears.
  '''
  is_reference = True
  def __init__(self, referent, orientation='n', surface=None, x=-1, y=-1):
    Fraxel.__init__(self, surface, x, y)
    self.referent = referent
    self.orientation = orientation

  def __hash__(self):
    return Fraxel.__hash__(self) - hash(self.referent) + ord(self.orientation)

  def __eq__(self, other):
    if type(other) != Reference:
      return False
    if self.surface == other.surface\
    and self.x == other.x\
    and self.y == other.y\
    and self.referent == other.referent\
    and self.orientation == other.orientation:
      return True
    return False

  def __ne__(self, other):
    return not self == other

  def __str__(self):
    if self.surface:
      return "a Reference to '{}' (orientation {}) at ({}, {}) on '{}'".format(
        self.referent, self.orientation,
        self.x, self.y, self.surface.name,
      )
    else:
      return "a Reference to '{}' (orientation {})".format(
        self.referent,
        self.orientation,
      )

  def __repr__(self):
    return "Reference(" + repr(self.referent) + ", " +\
                          repr(self.orientation) + ", " +\
                          repr(self.surface) + ", " +\
                          repr(self.x) + ", " + repr(self.y) + ")"

  def is_secondary(self):
    return self.surface.is_secondary(self.x, self.y)

  def update(self):
    if self.is_secondary():
      raise TypeError("Can't update a secondary Reference.")
    super().update()

  @property
  def root(self):
    return self.surface.find_root_of(self)

class Position(fractal.Position):
  '''
  An extension of a fractal Position that adds several convenient methods.
  '''
  def __init__(self, x, y, surface, parent=None):
    if parent:
      super().__init__(x, y, surface, parent)
    else:
      super().__init__(x, y, surface)

  __hash__ = None # TODO: something about this?

  def __str__(self):
    return "a Position at ({}, {}) on '{}'".format(
      self.x,
      self.y,
      self.surface.name,
    )

  def get_parent(self):
    '''
    Returns the parent of this position: either the real parent, or a new,
    made-up parent.
    '''
    if not self.parent:
      self.parent = Surface.find_reference_to(self.surface)
    return self.parent

  def ensure_heritage(self):
    '''
    Makes sure that this position has enough ancestors to be rendered.
    '''
    size_required = fractal.get_view_size()
    size_x = 1
    size_y = 1
    fixing = self
    while fixing and (size_x < size_required or size_y < size_required):
      size_x *= fixing.surface.hz_scale
      size_y *= fixing.surface.vt_scale
      fixing = fixing.get_parent()

  @property
  def fraxel(self):
    '''
    Returns the Fraxel that this position points to.
    '''
    return self.surface.get(self.x, self.y)

  @property
  def lineage(self):
    '''
    Returns the number of parents that this position has.
    '''
    this = self
    result = 0
    while this.parent:
      result += 1
      this = this.parent
    return result

class Surface(fractal.Surface):
  '''
  An extension of a fractal Surface that adds several convenient methods. This
  class also runs HSV<->RGB conversion, so it effectively operates in [0-1) HSV
  space. It also keeps track of a name and world for this surface.
  '''
  world = None
  name = "<nameless>"
  def __init__(
    self,
    name,
    world,
    sc_width=1,
    sc_height=1,
    hz_scale=5,
    vt_scale=5,
  ):
    if not isinstance(world, dict):
      raise TypeError("World must be some kind of dictionary.")
    if re.search('\s', name):
      raise ValueError("Surface names may not contain whitespace.")
    check_name(name)
    super().__init__(hz_scale, vt_scale, sc_width, sc_height)
    world[name] = self
    self.name = name
    self.world = world

  def __str__(self):
    result = self.name + ":\n"
    for y in range(self.height):
      for x in range(self.width):
        frx = self.get(x, y)
        if frx.is_reference:
          result += '?'
        elif frx.is_pixel:
          if (frx.v == 0):
            result += ' '
          elif (frx.s == 0):
            result += '+'
          else:
            result += '#'
      result += '\n'
    return result

  @property
  def width(self):
    return self.sc_width * self.hz_scale

  @property
  def height(self):
    return self.sc_height * self.vt_scale

  @staticmethod
  def load(
    sname,
    world,
    data,
    get_data=no_loader,
    comments=re.compile(r"^\s*#.*$", flags=re.M),
  ):
    '''
    Loads a surface from the given data, and puts it into the given world. This
    function then returns a list of unresolved references (Fraxels with a
    string instead of a Surface as referents). It also takes a get_data
    function that should return data from the world directory given a filename
    within it (the other parameter allows customization of the definition of
    comments; it should be a regex). First, comments (lines starting with '#'
    as the first non-whitespace character) are stripped. Then aliases are read
    from the top of the file until a line containing just '---' is reached (or
    a bad alias shows up, in which case an exception is raised). The format for
    aliases is a single non-colon character, followed by a colon, followed by
    an optional space and then either the name of a surface (with an optional
    orientation specifier ('/n', '/e', '/s', or '/w') after it) or a color as
    '%<h>,<s>,<v>' (each of which should be a decimal) or '~<H>,<S>,<V>' (each
    of which should be an integer). After the separator, the next line defines
    the dimensions of the surface in scale units (i.e. its size when referred
    to) using '<width>x<height>'. Then if the next line starts with a colon,
    the rest of the line is a filename containing an image that defines pixel
    values. Otherwise, the following lines should contain characters that
    describe the pixel values in terms of the aliases defined above. Finally,
    there may be an additional separator ('---' on its own line), after which
    there may follow any number of references, each specified as
    '@<x>,<y>:<name>/<orientation>' where the orientation is optional (it
    defaults to '/n').
    '''
    data = comments.sub("", data)
    aliasmap = {}
    parsing = 'aliases'
    first = True
    second = True
    sc_width, sc_height = 1, 1
    width, height = None, 0
    rows = []
    for line in data.split('\n'):
      if parsing == 'aliases':
        if not line.strip():
          continue
        if line == '---':
          parsing = 'data'
          continue
        key = line[0]
        val = line[2:].strip()
        if val[0] in '%~':
          color = colors.Color.parse(val)
          debug.debug(
            "unpack",
            "Found pixel alias '{}' for {}".format(key, color)
          )
          aliasmap[key] = Pixel.colored(color)
        else:
          orientation = 'n'
          if '/' in val:
            name, orientation = val.split('/')
          else:
            name = val
          if name in ("<name>", "<unknown>"):
            raise ValueError("Surface name '{}' is reserved!".format(name))
          if name in world:
            referent = (world[name], orientation)
            debug.debug(
              "unpack",
              "Found surface alias '{}' for '{}'.".format(key, referent[0].name)
            )
          else:
            referent = (name, orientation)
            debug.debug(
              "unpack",
              "Found name alias '{}'.".format(key, referent[0])
            )
          aliasmap[key] = Reference(*referent)
      elif parsing == 'data':
        if line == '':
          continue
        if line == '---':
          parsing = 'data'
          continue
        if first:
          first = False
          sc_width, sc_height = line.strip().split('x')
          sc_width, sc_height = int(sc_width), int(sc_height)
          continue
        elif second:
          second = False
          if line[0] == ':':
            filename = line.strip()[1:]
            data = get_data(filename)
            # TODO: SOMETHING HERE!
            raise NotImplementedError(
              "File-based surface definitions aren't implemented yet."
            )
            continue
        height += 1
        rows.append([])
        if width == None:
          width = len(line)
        elif len(line) != width:
          raise ValueError("Surface data doesn't have uniform width!")
        for chr in line:
          if chr not in aliasmap:
            raise ValueError("Unmapped character '{}'.".format(chr))
          rows[-1].append(aliasmap[chr])
      elif parsing == 'references':
        if not line.strip():
          continue
        if line == '---':
          parsing = None
          continue
        loc, rest = line.strip().split(':')
        x, y = loc.split(',')
        x = x[1:]
        x, y = int(x), int(y)
        orientation = 'n'
        rest = rest.strip()
        if '/' in rest:
          name, orientation = rest.split('/')
        else:
          name = rest
        if name in ("<name>", "<unknown>"):
          raise ValueError("Surface name '{}' is reserved!".format(name))
        if name in world:
          referent = (world[name], orientation)
        else:
          referent = (name, orientation)
        rows[y][x] = Reference(*referent)
      else:
        break
    hz_scale = width // sc_width
    vt_scale = height // sc_height
    surface = Surface(sname, world, sc_width, sc_height, hz_scale, vt_scale)
    unresolved = []
    for y, row in enumerate(rows):
      for x, val in enumerate(row):
        if val.is_reference:
          if isinstance(val.referent, Surface):
            cmp = surface.get(x, y)
            if not cmp.is_reference\
            or cmp.referent != val.referent\
            or cmp.orientation != val.orientation:
                surface.set(x, y, val)
          else:
            unresolved.append(
              Reference(val.referent, val.orientation, surface, x, y)
            )
        elif val.is_pixel:
          surface.set(x, y, val)
    return unresolved

  @staticmethod
  def find_reference_to(surface):
    '''
    Searches through the surface's world and finds the first Surface that
    references it. Returns a Position that points to the upper-left square of
    the target Surface within the found Surface, or None if there are no
    references to the target, or if the target doesn't have a world.
    '''
    if not hasattr(surface, "world"):
      return None
    for key in surface.world:
      if key == '<name>':
        continue
      s = surface.world[key]
      for y in range(s.height):
        for x in range(s.width):
          frx = s.get(x, y)
          if frx.is_reference and frx.referent == surface:
            return Position(x, y, s)
    return None

  @staticmethod
  def find_root_of(fraxel):
    '''
    Takes a Fraxel and returns a Fraxel that's the root of the given Fraxel (or
    None if it has no root). The root of a Fraxel is the primary reference for
    which that Fraxel is a secondary reference.
    '''
    result = fraxel.surface.find_root(fraxel.x, fraxel.y)
    if result:
      return fraxel.surface.get(result[0], result[1]);
    return None

  def get(self, x, y):
    '''
    Returns a Fraxel from the Surface. If the Fraxel is a Reference, the
    referent returned may not be referenced from the exact coordinates given;
    it might be a secondary reference. Use is_secondary to check. Note that
    out-of-bounds coordinates will return Edge Fraxels, and negative
    coordinates are out-of-bounds. The orientation of a returned Reference will
    be one of 'n', 'e', 's', or 'w'.
    '''
    if x < 0 or y < 0:
      return Edge(self, x, y)
    value = super().get(x, y)
    if value and not isinstance(value[0], fractal.Surface):
      value = colorsys.rgb_to_hsv(*[val/255. for val in value])
      return Pixel(value[0], value[1], value[2], surface=self, x=x, y=y)
    if value:
      value = list(value)
      if value[1] == 0:
        value[1] = 'n'
      elif value[1] == 1:
        value[1] = 'e'
      elif value[1] == 2:
        value[1] = 's'
      elif value[1] == 3:
        value[1] = 'w'
      return Reference(value[0], value[1], self, x, y)
    return Edge(self, x, y)

  def set(self, x, y, fraxel):
    '''
    Changes the surface by setting the fraxel at the given coordinates to the
    given Fraxel. The value should be either a Pixel or a Reference (Edges
    can't be part of a Surface proper). Note that neighboring fraxels will be
    overwritten according to the size of the Surface. Note also that if the
    destination value was a Surface, all fraxels that were part of the
    reference to it will be cleared to black. For placing Surfaces in various
    orientations, the coordinates given here are always the upper-left corner
    of the placed reference within the surface being changed. 
    '''
    ori = 0
    if isinstance(fraxel, colors.Color):
      fraxel = Pixel.colored(fraxel)
    elif isinstance(fraxel, Surface):
      fraxel = Reference(fraxel)
    if fraxel.is_pixel:
      super().set(
        x,
        y,
        [int(255*val)
         for val in colorsys.hsv_to_rgb(fraxel.h, fraxel.s, fraxel.v)]
      )
    elif fraxel.is_reference:
      if fraxel.orientation in ("east", "e", "E"):
        ori = 1
      elif fraxel.orientation in ("south", "s", "S"):
        ori = 2
      elif fraxel.orientation in ("west", "w", "W"):
        ori = 3
      super().set(x, y, fraxel.referent, ori)
    else:
      raise TypeError("Edge Fraxels can't be put into Surfaces.")

  def fill(self, pixel):
    '''
    Fills the surface with the given color.
    '''
    if isinstance(pixel, colors.Color):
      pixel = Pixel.colored(pixel)
    if not pixel.is_pixel:
      raise ValueError(
        "Can't use a {} as a border color.".format(type(pixel))
      )
    for x in range(self.width):
      for y in range(self.height):
        self.set(x, y, pixel)

  def set_border(self, pixel):
    '''
    Adds a border to the surface, setting each of the edge pixels to the given
    color.
    '''
    if isinstance(pixel, colors.Color):
      pixel = Pixel.colored(pixel)
    if not pixel.is_pixel:
      raise ValueError(
        "Can't use a {} as a border color.".format(type(pixel))
      )
    for x in range(self.width):
      for y in range(self.height):
        if x == 0 or y == 0 or x == self.width - 1 or y == self.height - 1:
          self.set(x, y, pixel)

  def obliterate(self):
    '''
    Removes this surface from its world, and sets all references to it in other
    surfaces in its world to (0, 0, 0), allowing it to be garbage collected
    (assuming other references to it are gone). Note that self-references
    within the surface are not cleared (TODO: is this true?).
    '''
    del self.world[self.name]
    for key in self.world:
      if key == '<name>':
        continue
      s = self.world[key]
      for x in range(s.width):
        for y in range(s.height):
          test = s.get(x, y)
          if test.is_reference and test.referent == self:
            s.set(x, y, Pixel(0, 0, 0))
