#!/usr/bin/env python3
import os, re, sys, time, warnings, zipfile

from types import FunctionType as function

from collections import Iterable as iterable

from .lib import symbiosis, fractal

from . import colors, frwrap, engine

from . import debug

###########
# Globals #
###########

RULES = None

KEYMAP = {}

VIEW_HISTORY = {
  'age': 1,
  'position': None,
  'angle': 0,
  'hz_scale': 1,
  'vt_scale': 1,
}

LAST_STEP_TIME = 0

ZOOM = 0

# TODO: Make this accessible (at least to toponaut.py)
TIME_PER_STEP = 0.2

KEYS = set()
TRIGGERED = set()

#####################
# Loading functions #
#####################

def load_keymap(name):
  '''
  Loads the named keymap (from the 'name.map' file).
  '''
  # TODO: Make this non-dangerous.
  locals = {}
  with open(os.path.normcase(name) + ".map", 'rU') as fin:
    code = fin.read()
    exec(code, globals(), locals)
  km = dict(locals["keymap"])
  result = {}
  for k, binding in km.items():
    if type(k) == bytes:
      result[k] == binding
    if type(k) == str:
      byt = bytes(k, "utf-8")
      if byt not in result:
        result[byt] = binding
  return result

def set_keymap(km):
  '''
  Sets the active key mapping.
  '''
  global KEYMAP
  KEYMAP = km

def load_rules(name, in_progress=set()):
  '''
  Takes the name of a ruleset and loads it. This involves reading the
  appropriate .rls file, and then executing the contents of it in a fertile
  environment, which is dangerous (the rls file could contain all manner of
  hacks or viruses).
  '''
  if name in in_progress:
    raise ValueError(
      "Can't load rule set '{}' while it is already being loaded.".format(name)
    )
  # TODO: Make this safer?
  filename = os.path.join("rules", os.path.normcase(name) + '.rls')
  if not os.path.isfile(filename):
    raise ValueError(
      "Ruleset '{}' not found (no '{}' file).".format(name, filename)
    )
  with open(filename, 'rU') as fin:
    data = fin.read()
  env = get_env()
  bad = False
  in_progress.add(name)
  try:
    instructions = compile(data, "<ruleset '{}'>".format(name), "exec")
    exec(instructions, env)
  except Exception:
    print(
      "The following error occurred during rule loading:\n",
      file=sys.stderr
    )
    sys.excepthook(*sys.exc_info())
    print("\n...which caused the loading process to fail.\n", file=sys.stderr)
    bad = True
  in_progress.remove(name)
  if bad:
    raise ValueError(
      "Ruleset '{}' failed to load (error during loading).".format(name)
    )
  if 'RULES' not in env:
    raise ValueError(
      "Ruleset '{}' failed to load (didn't define 'RULES').".format(name)
    )
  if not isinstance(env['RULES'], Ruleset):
    raise ValueError(
      "Ruleset '{}' failed to load ('RULES' wasn't a Ruleset).".format(name)
    )
  rs = env['RULES']
  if rs.name == "<unnamed>":
    rs.name = name
  return rs

def load_world(worldname):
  '''
  Loads a fractal world by finding either a directory named <world> or zip file
  named <world>.zip in the 'worlds' directory. Returns the resulting world
  dictionary, plus a list of Fraxels that couldn't be resolved within that
  world. If the world is found as a directory, it should contain a
  'surfaces.list' file (literal 'world') that lists surfaces and a starting
  position. If it's a zip file, it should unzip into a directory that would
  contain a 'surfaces.list' file. The world file should list one surface per
  line, giving just the name (lines starting with '#' are comments, and blank
  lines are ignored). Each surface is loaded from a file named <name>.frs in
  the same directory (or zipped directory) as the world file.
  '''
  unresolved = []

  world = engine.make_world(os.path.split(worldname)[-1])

  worldfile = os.path.normcase(os.path.join("worlds", worldname))
  if os.path.isdir(worldfile):
    listfile = os.path.join(worldfile, 'surfaces.list')
    if not os.path.isfile(listfile):
      raise ValueError("World '{}' has no surfaces list.".format(worldname))
    with open(listfile, 'rU') as fin:
      for line in fin:
        name = line.strip()
        if not name or name[0] == '#':
          continue
        if name == '<name>':
          raise ValueError("Surface name '<name>' is reserved!")
        sfile = os.path.join(worldfile, os.path.normcase(name) + ".frs")
        if not os.path.isfile(sfile):
          raise ValueError(
            "Surface '{}' is missing (file not found).".format(name)
          )
        with open(sfile, 'rU') as sfin:
          data = sfin.read()
        try:
          leftovers = frwrap.Surface.load(name, world, data)
        except:
          sys.excepthook(*sys.exc_info())
          print(
            "Error while loading surface '{}' from world '{}'.".format(
              name,
              world['<name>']
            ),
            file=sys.stderr
          )
          exit(1)
        if leftovers:
          unresolved.extend(leftovers)
  elif zipfile.is_zipfile(worldfile + '.zip'):
    zf = zipfile.ZipFile(worldfile + '.zip')
    surfaces = zf.read(world['<name>'] + '/surfaces.list', 'rU')
    surfaces = surfaces.decode('utf-8')
    for line in surfaces.split('\n'):
      name = line.strip()
      if not name or name[0] == '#':
        continue
      if name in ("<name>", "<unknown>"):
        raise ValueError("Surface name '{}' is reserved!".format(name))
      data = zf.read(world['<name>'] + '/' + name + '.frs')
      data = data.decode('utf-8')
      try:
        leftovers = frwrap.Surface.load(name, world, data)
      except:
        sys.excepthook(*sys.exc_info())
        print(
          "Error while loading surface '{}' from world '{}'.".format(
            name,
            world['<name>']
          ),
          file=sys.stderr
        )
        exit(1)
      if leftovers:
        unresolved.extend(leftovers)
    zf.close()
  else:
    raise ValueError("World '{}' could not be found.".format(worldname))
  leftovers = []
  for frx in unresolved:
    if frx.referent in world:
      frx.referent = world[frx.referent]
      cmp = frx.surface.get(frx.x, frx.y)
      if not cmp.is_reference\
      or cmp.referent != frx.referent\
      or cmp.orientation != frx.orientation:
        try:
          cmp.become(frx)
        except:
          sys.excepthook(*sys.exc_info())
          print(
            "Error while resolving reference to '{}' on surface '{}' in "
            "world '{}'.".format(
              frx.referent.name,
              cmp.surface.name,
              cmp.surface.world['<name>']
            ),
            file=sys.stderr
          )
          exit(1)
    elif not isinstance(frx.referent, frwrap.Surface):
      leftovers.append(frx)
      frx.become(frwrap.Reference(world["<unknown>"], 'n'))
  return world, leftovers

def load_universe(start):
  '''
  Takes a world name, and loads that world, and then all worlds referenced by
  it, and so on and so forth, until it runs out of worlds to load (or the
  program runs out of RAM and crashes). Returns a dictionary of worlds indexed
  by their names. Note that
  '''
  universe = {}
  leftovers = []
  unknown_worlds = [start]
  debug.debug("references", "---Starting to load a universe.---")
  while unknown_worlds:
    next = unknown_worlds.pop()
    world, leftover = load_world(next)
    universe[world['<name>']] = world
    leftovers.extend(leftover)
    still_unresolved = []
    # Resolve what leftovers we can, and discover any worlds that need loading:
    for frx in leftovers:
      debug.debug("references", "Leftover reference:", frx)
      if '.' in frx.referent:
        wld, srf = frx.referent.split('.')
        if wld in universe and srf in universe[wld]:
          frx.referent = universe[wld][srf]
          debug.debug("references", " ...resolved!")
        elif wld in universe:
          frx.referent = frx.surface.world['<unknown>']
          debug.debug(
            "references",
            " ...couldn't be resolved (unknown surface)."
          )
        else:
          unknown_worlds.append(wld)
          still_unresolved.append(frx)
          debug.debug("references", " ...found a new world: '{}'!".format(wld))
          continue
      elif frx.referent in frx.surface.world:
        frx.referent = frx.surface.world[frx.referent]
        debug.debug("references", " ...by mistake (fixed)!".format(wld))
      cmp = frx.surface.get(frx.x, frx.y)
      if not cmp.is_reference\
      or cmp.referent != frx.referent\
      or cmp.orientation != frx.orientation:
        try:
          cmp.become(frx)
        except:
          sys.excepthook(*sys.exc_info())
          print(
            "Error while resolving reference to '{}' on surface '{}' in "
            "world '{}'.".format(
              frx.referent.name,
              cmp.surface.name,
              cmp.surface.world['<name>']
            ),
            file=sys.stderr
          )
          exit(1)
    leftovers = still_unresolved
  debug.debug("references", "---No more worlds to find.---")
  # Resolve any leftover leftovers...
  for frx in leftovers:
    debug.debug("references", "Leftover reference:", frx)
    if '.' in frx.referent:
      wld, srf = frx.referent.split('.')
      if wld in universe and srf in universe[wld]:
        frx.referent = universe[wld][srf]
        debug.debug("references", " ...resolved!")
      else:
        frx.referent = frx.surface.world['<unknown>']
        debug.debug(
          "references",
          " ...couldn't be resolved (unknown world or surface)."
        )
    elif frx.referent in frx.surface.world:
      debug.debug("references", " ...by mistake (fixed)!".format(wld))
      frx.referent = frx.surface.world[frx.referent]
    elif not isinstance(frx.referent, frwrap.Surface):
      debug.debug("references", " ...odd value (set to unknown).".format(wld))
      frx.referent = frx.surface.world['<unknown>']
    cmp = frx.surface.get(frx.x, frx.y)
    if not cmp.is_reference\
    or cmp.referent != frx.referent\
    or cmp.orientation != frx.orientation:
      cmp.become(frx)
  debug.debug("references", "---Done sorting out references.---")
  return universe

def get_mission_groups():
  '''
  Returns the mission list, which is a dictionary of mission group names each
  of which indexes a list of mission names.
  '''
  files = os.listdir(os.path.normcase("missions"))
  groups = {}
  for f in files:
    missions = []
    groupdir = os.path.normcase(os.path.join("missions", f))
    if os.path.isdir(groupdir):
      mfiles = os.listdir(groupdir)
      for m in mfiles:
        if m.endswith(".frm"):
          missions.append(m[:-4])
    if missions:
      groups[f] = missions
  return groups

####################
# Helper functions #
####################

def zoom_out():
  set_zoom(ZOOM + 0.2)

def zoom_in():
  set_zoom(ZOOM - 0.2)

def rezoom():
  set_zoom(0)

def set_zoom(zoom):
  global ZOOM
  ZOOM = zoom
  fractal.set_zoom(ZOOM)
  ZOOM = fractal.get_zoom()

def quit():
  symbiosis.quit()

KEYMAP['q'] = quit
#KEYMAP[''] = quit

def get_rules():
  '''
  Returns the current Ruleset.
  '''
  global RULES
  return RULES

def set_rules(rules):
  '''
  Redefines the rules used to run the world. Takes a new Ruleset object. Note
  that the new setup rules of the new set won't ever get run unless further
  action is taken to do so.
  '''
  global RULES
  RULES = rules

def handle_input():
  '''
  Executes all of the actions mapped from keys either triggered or held when
  the call is made.
  '''
  global KEYS, TRIGGERED
  keyset = KEYS | TRIGGERED
  for key in sorted(list(keyset)):
    if key in KEYMAP:
      action = KEYMAP[key]
      if isinstance(action, function):
        action()
      elif isinstance(action, str):
        engine.raise_signals(action)
      elif isinstance(action, iterable):
        engine.raise_signals(*action)
  TRIGGERED = set()

def get_player():
  '''
  Looks up and returns the player's position and the Pixel at that position. If
  the player's position isn't defined, it returns (None, None), and if the
  player's position is defined, but there's no Pixel there (i.e. it's on an
  Edge or Reference), it returns the position found and None.
  '''
  ppos = engine.get_position("player")
  if not ppos:
    return None, None
  ppx = ppos.fraxel
  if not ppx.is_pixel:
    return ppos, None
  return ppos, ppx

def is_player(pixel):
  '''
  Returns True if the given Pixel is at the player's current position. Always
  returns false if the player's location isn't a Pixel or if the player's
  location is undefined.
  '''
  ppos, ppx = get_player()
  if not ppos or not ppx:
    return False
  return all([
    ppx.surface == pixel.surface,
    ppx.x == pixel.x,
    ppx.y == pixel.y
  ])

###########
# Classes #
###########

class Mission:
  '''
  A Mission is an organized chunk of playable content. Each mission consists of
  a starting location (a worldname, surfacename, x, y tuple) and a goal
  location (the same), along with a starting color for the player and the world
  (implying a universe) and ruleset that should be loaded. Each mission has a
  name, and the missions are organized into named groups. Missions can also
  require that other missions have been completed.
  '''
  def __init__(self, group, name, rules, start, goal, color, requires=None):
    '''
    Sets up the mission. Group and name should be strings, rules should be a
    Ruleset, start and goal should each be 4-tuples listing worldname,
    surfacename, x, and y in that order, and color should be a Pixel or a
    Color. Requires, if given, should be a list of "group.name" strings
    referring to other missions.
    '''
    frwrap.check_name(group)
    frwrap.check_name(name)
    self.group = group
    self.name = name
    self.rules = rules
    self.start_world, self.start_surface, self.start_x, self.start_y = start
    self.goal_world, self.goal_surface, self.goal_x, self.goal_y = goal
    if isinstance(color, colors.Color):
      self.start_pixel = frwrap.Pixel.colored(color)
    elif isinstance(color, frwrap.Pixel):
      self.start_pixel = color
    else:
      raise TypeError("Bad color (must be either a Color or a Pixel).")
    self.requires = requires or []

  @staticmethod
  def load(group, name):
    '''
    Loads a mission file and returns the resulting Mission object.
    '''
    rulesets = []
    requirements = []
    start = None
    goal = None
    color = None
    with open(
      os.path.normcase(os.path.join("missions", group, name + ".frm")),
      'rU'
    ) as fin:
      parsing = None
      for i, line in enumerate(fin):
        ln = i + 1
        line = line.strip()
        if line[0] == "#":
          continue
        if line == "rules:":
          parsing = "rules"
          continue
        if line == "requires:":
          parsing = "requires"
          continue
        elif line.startswith("start:"):
          parsing = None
          surface, pos = line[6:].strip().split(':')
          world, surf = surface.split('.')
          x, y = (int(v.strip()) for v in pos.strip("()").split(','))
          start = world, surf, x, y
        elif line.startswith("goal:"):
          parsing = None
          surface, pos = line[5:].strip().split(':')
          world, surf = surface.split('.')
          x, y = (int(v.strip()) for v in pos.strip("()").split(','))
          goal = world, surf, x, y
        elif line.startswith("color:"):
          parsing = None
          color = colors.Color.parse(line[6:])
        elif parsing == "rules":
          rulesets.append(line)
        elif parsing == "requires":
          requirements.append(line)
        else:
          raise ValueError(
            "Unidentified line (#{}) in mission file.".format(ln)
          )

    rules = Ruleset()
    for rs in rulesets:
      rules |= load_rules(rs)

    if not start:
      raise ValueError("Couldn't load mission (no starting location.")
    elif not goal:
      raise ValueError("Couldn't load mission (no goal location.")
    elif not color:
      raise ValueError("Couldn't load mission (no starting color.")

    return Mission(group, name, rules, start, goal, color, requirements)

class Ruleset:
  '''
  A Ruleset is an organized set of rules: it contains lists of setup and normal
  rules, and it also has a list of helper functions. Rulesets can be loaded
  from .rls files by the load_rules function. Rulesets also support the set
  union operation ('|') for constructing new rulesets. Note that unlike a true
  intersection operation, intersecting rulesets can result in duplicated rules.
  '''
  def __init__(
    self,
    setuprules=None,
    normalrules=None,
    helpers=None,
    name="<unnamed>",
  ):
    frwrap.check_name(name)
    self.name = name
    self.setuprules = setuprules or []
    self.normalrules = normalrules or []
    self.helpers = helpers or []
    self.names = {}
    for r in self.setuprules:
      self.names[r.name] = r
    for r in self.normalrules:
      self.names[r.name] = r
    for f in self.helpers:
      self.names[f.__code__.co_name] = f
    for reserved in [
      'name',
      'setuprules',
      'normalrules',
      'helpers',
      'names',
      '__init__',
      '__getattr__',
    ]:
      if reserved in self.names:
        warnings.warn(
          "Rule '{}' has a reserved name which won't be visible.".format(
            reserved
          )
        )

  def __getattr__(self, attr):
    if attr in self.names:
      return self.names[attr]

  def __or__(self, other):
    if not isinstance(other, Ruleset):
      raise TypeError("Rulesets can only be unified with other Rulesets.")
    setuprules = self.setuprules + other.setuprules
    normalrules = self.normalrules + other.normalrules
    helpers = self.helpers + other.helpers
    return Ruleset(setuprules, normalrules, helpers)

###################################
# The rule definition environment #
###################################

def get_env(ENV={}):
  '''
  Returns an environment suitable for setting up a rules file in.
  '''
  if not ENV:
    # Make all Fraxel subclasses available:
    frns = frwrap.__dict__
    for name in frns:
      if type(frns[name]) == type and issubclass(frns[name], frwrap.Fraxel):
        ENV[name] = frns[name]
    # The list of available functions:
    available = [
      quit,
      load_universe, engine.set_universe,
      load_world, engine.make_world, engine.add_world, engine.get_world,
      engine.reset,
      get_mission_groups,
      engine.find_surface,
      load_keymap, set_keymap,
      load_rules, get_rules, set_rules,
      get_player, is_player,
      zoom_in, zoom_out, rezoom, set_zoom,
      engine.set_biases, engine.set_tilt,
      engine.dirint, engine.dirchar,
      engine.get_step,
      engine.get_value, engine.set_value, engine.clear_values,
      engine.get_position, engine.set_position, engine.clear_positions,
      engine.check_signals, engine.get_signals, engine.raise_signals,
      engine.clear_signals,
      engine.move, engine.fix_positions,
      engine.get_neighbor,
      engine.rule,
      engine.apply_rules,
    ]
    # ...plus all of the functions from the colors module:
    for name in dir(colors):
      f = getattr(colors, name)
      if isinstance(f, function):
        available.append(f)
    # Add the available functions to the environment:
    for f in available:
      ENV[f.__code__.co_name] = f
    # And then add a few extra classes plus some pre-defined colors:
    ENV["FORBIDDEN_CHARACTERS"] = frwrap.FORBIDDEN_CHARACTERS
    ENV["Ruleset"] = Ruleset
    ENV["Mission"] = Mission
    ENV["Rule"] = engine.Rule
    ENV["Color"] = colors.Color
    ENV["Position"] = frwrap.Position
    ENV["Surface"] = frwrap.Surface
    ENV["BLACK"] = colors.Color(0, 0, 0)
    ENV["GREY"] = colors.Color(0, 0, 0.5)
    ENV["WHITE"] = colors.Color(0, 0, 1)
    ENV["RED"] = colors.Color(hue="red", s=1, v=1)
    ENV["ORANGE"] = colors.Color(hue="orange", s=1, v=1)
    ENV["YELLOW"] = colors.Color(hue="yellow", s=1, v=1)
    ENV["LIME"] = colors.Color(hue="lime", s=1, v=1)
    ENV["GREEN"] = colors.Color(hue="green", s=1, v=1)
    ENV["TEAL"] = colors.Color(hue="teal", s=1, v=1)
    ENV["CYAN"] = colors.Color(hue="cyan", s=1, v=1)
    ENV["AZURE"] = colors.Color(hue="azure", s=1, v=1)
    ENV["BLUE"] = colors.Color(hue="blue", s=1, v=1)
    ENV["PURPLE"] = colors.Color(hue="purple", s=1, v=1)
    ENV["MAGENTA"] = colors.Color(hue="magenta", s=1, v=1)
    ENV["PINK"] = colors.Color(hue="pink", s=1, v=1)
  return dict(ENV)

#############
# Callbacks #
#############

def resize(w, h):
  fractal.resize(w, h)

def render():
  global RULES, VIEW_HISTORY, LAST_STEP_TIME, TIME_PER_STEP
  now = time.time()
  ppos = engine.get_position("player")
  if now - LAST_STEP_TIME > TIME_PER_STEP and ppos:
    VIEW_HISTORY['age'] = 0
    VIEW_HISTORY['position'] = ppos
    handle_input()
    engine.step(
      RULES.normalrules,
      ppos.surface.world,
    )
    ppos = engine.get_position("player")
    last_pos = VIEW_HISTORY['position']
    if last_pos:
      pplin = ppos.lineage
      lastlin = last_pos.lineage
      # TODO: Sort these out...
      if pplin == lastlin + 1\
      and ppos.get_parent().fraxel.is_reference:
        VIEW_HISTORY['angle'] = engine.dirint(
          ppos.get_parent().fraxel.orientation
        ) * 90
        VIEW_HISTORY['hz_scale'] = 1/ppos.surface.hz_scale
        VIEW_HISTORY['vt_scale'] = 1/ppos.surface.vt_scale
      elif pplin == lastlin - 1\
      and last_pos.get_parent().fraxel.is_reference:
        VIEW_HISTORY['angle'] = engine.dirint(
          last_pos.get_parent().fraxel.orientation
        ) * 90
        VIEW_HISTORY['hz_scale'] = last_pos.surface.hz_scale
        VIEW_HISTORY['vt_scale'] = last_pos.surface.vt_scale
      elif pplin == lastlin\
      and ppos.surface != last_pos.surface:
        this_ori = 'n'
        last_ori = 'n'
        this_parent = ppos.get_parent()
        last_parent = last_pos.get_parent()
        if this_parent:
          this_ori = this_parent.fraxel.orientation
        if last_parent:
          last_ori = last_parent.fraxel.orientation
        VIEW_HISTORY['angle'] = (
          engine.dirint(
            last_ori
          ) - engine.dirint(
            this_ori
          )
        ) * 90
        VIEW_HISTORY['hz_scale'] = last_pos.surface.hz_scale /\
                                   float(ppos.surface.hz_scale)
        VIEW_HISTORY['vt_scale'] = last_pos.surface.vt_scale /\
                                   float(ppos.surface.vt_scale)
      # TODO: Are there any other cases to address?
      else:
        VIEW_HISTORY['angle'] = 0
        VIEW_HISTORY['hz_scale'] = 1
        VIEW_HISTORY['vt_scale'] = 1
    LAST_STEP_TIME = now
  VIEW_HISTORY['age'] += 0.01 # TODO: adjust this.
  symbiosis.clear_color_buffer()
  if ppos:
    ppos.ensure_heritage()
    fractal.render(
      ppos,
      float(VIEW_HISTORY['angle']),
      float(VIEW_HISTORY['hz_scale']),
      float(VIEW_HISTORY['vt_scale']),
      float(VIEW_HISTORY['age'])
    )
  return True

def keyboard(pressed, key, x, y):
  global KEYS, TRIGGERED
  if pressed:
    debug.debug("input", "Key pressed: '{}'".format(key.decode()))
    KEYS.add(key)
    TRIGGERED.add(key)
  else:
    debug.debug("input", "Key released: '{}'".format(key.decode()))
    KEYS.discard(key)
    KEYS.discard(key.upper())
    KEYS.discard(key.lower())

#########
# Setup #
#########

callbacks = {
  "render": render,
  "resize": resize,
  "keyboard": keyboard,
}

#################
# Main function #
#################

def start(ruleset="boot", keymap="keys", frame_time=0.2):
  global RULES, TIME_PER_STEP
  TIME_PER_STEP = frame_time
  set_keymap(load_keymap(keymap))
  set_rules(load_rules(ruleset))
  engine.apply_rules(RULES.setuprules)
  title = "Resemble"
  wt = engine.get_value("window_title")
  if wt:
    title = "Resemble - {}".format(wt)
  symbiosis.prepare(
    width=800,
    height=600,
    name=title,
    background=(0.5, 0.5, 0.5, 1),
  )
  symbiosis.attach(callbacks)
  symbiosis.loop()
