#include "Python.h"

#include <GL/gl.h>
#include <GL/glut.h>

#include <malloc.h>
#include <stdlib.h>

static PyObject* idleCB = NULL;
static PyObject* resizeCB = NULL;
static PyObject* renderCB = NULL;
static PyObject* keyboardCB = NULL;
static PyObject* specialCB = NULL;
static PyObject* mouseCB = NULL;
static PyObject* motionCB = NULL;

static int WINDOW;

static const int DEFAULT_MODE = GLUT_RGBA | GLUT_DOUBLE;
static const int DEFAULT_WIDTH = 800;
static const int DEFAULT_HEIGHT = 600;

// TODO: Check Py_BuildValue running out of memory.
// TODO: Better/cleaner exit methods?

// Forward declarations:

static void fail(void);

// C Callback functions:

static void idle(void) {
  PyObject* ret;
  if (idleCB != NULL && PyCallable_Check(idleCB)) {
    ret = PyObject_CallObject(idleCB, NULL);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    } else if (PyObject_IsTrue(ret)) {
      glutPostRedisplay();
    }
    Py_XDECREF(ret);
  } else {
    glutPostRedisplay();
  }
}

static void resize(int w, int h) {
  PyObject *args, *ret;
  if (resizeCB != NULL && PyCallable_Check(resizeCB)) {
    args = Py_BuildValue("(ii)", w, h);
    ret = PyObject_CallObject(resizeCB, args);
    if (PyObject_IsTrue(ret)) {
      glViewport(0, 0, w, h);
    }
    Py_XDECREF(ret);
    Py_XDECREF(args);
  } else {
    glViewport(0, 0, w, h);
  }
}

static void render(void) {
  PyObject* ret;
  if (renderCB != NULL && PyCallable_Check(renderCB)) {
    ret = PyObject_CallObject(renderCB, NULL);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    } else if (PyObject_IsTrue(ret)) {
      glutSwapBuffers();
      glClear( GL_COLOR_BUFFER_BIT );
    }
    Py_XDECREF(ret);
  } else {
    glutSwapBuffers();
    glClear( GL_COLOR_BUFFER_BIT );
  }
}

static void keyboard(unsigned char key, int x, int y) {
  PyObject *args, *ret;
  if (keyboardCB != NULL && PyCallable_Check(keyboardCB)) {
    args = Py_BuildValue("(Ocii)", Py_True, (int) key, x, y);
    ret = PyObject_CallObject(keyboardCB, args);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    }
    Py_XDECREF(ret);
    Py_XDECREF(args);
  }
}

static void keyboardUp(unsigned char key, int x, int y) {
  PyObject *args, *ret;
  if (keyboardCB != NULL && PyCallable_Check(keyboardCB)) {
    args = Py_BuildValue("(Ocii)", Py_False, (int) key, x, y);
    ret = PyObject_CallObject(keyboardCB, args);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    }
    Py_XDECREF(ret);
    Py_XDECREF(args);
  }
}

static void special(int key, int x, int y) {
  PyObject *args, *ret;
  if (specialCB != NULL && PyCallable_Check(specialCB)) {
    args = Py_BuildValue("(Oiii)", Py_True, key, x, y);
    ret = PyObject_CallObject(specialCB, args);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    }
    Py_XDECREF(ret);
    Py_XDECREF(args);
  }
}

static void specialUp(int key, int x, int y) {
  PyObject *args, *ret;
  if (specialCB != NULL && PyCallable_Check(specialCB)) {
    args = Py_BuildValue("(Oiii)", Py_False, key, x, y);
    ret = PyObject_CallObject(specialCB, args);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    }
    Py_XDECREF(ret);
    Py_XDECREF(args);
  }
}

static void mouse(int button, int state, int x, int y) {
  PyObject *args, *ret;
  if (mouseCB != NULL && PyCallable_Check(mouseCB)) {
    if (state == GLUT_DOWN) {
      args = Py_BuildValue("(Oiii)", Py_True, button, x, y);
    } else {
      args = Py_BuildValue("(Oiii)", Py_False, button, x, y);
    }
    ret = PyObject_CallObject(mouseCB, args);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    }
    Py_XDECREF(ret);
    Py_XDECREF(args);
  }
}

static void motion(int x, int y) {
  PyObject *args, *ret;
  if (motionCB != NULL && PyCallable_Check(motionCB)) {
    args = Py_BuildValue("(ii)", x, y);
    ret = PyObject_CallObject(motionCB, args);
    if (ret == NULL) {
      PyErr_Print();
      fail();
    }
    Py_XDECREF(ret);
    Py_XDECREF(args);
  }
}

// Callback management functions:

static void attach(PyObject* callbacks) {
  if (PyMapping_Check(callbacks)) {
    if (PyMapping_HasKeyString(callbacks, "idle")) {
      idleCB = PyMapping_GetItemString(callbacks, "idle");
    }
    if (PyMapping_HasKeyString(callbacks, "resize")) {
      resizeCB = PyMapping_GetItemString(callbacks, "resize");
    }
    if (PyMapping_HasKeyString(callbacks, "render")) {
      renderCB = PyMapping_GetItemString(callbacks, "render");
    }
    if (PyMapping_HasKeyString(callbacks, "keyboard")) {
      keyboardCB = PyMapping_GetItemString(callbacks, "keyboard");
    }
    if (PyMapping_HasKeyString(callbacks, "special")) {
      specialCB = PyMapping_GetItemString(callbacks, "special");
    }
    if (PyMapping_HasKeyString(callbacks, "mouse")) {
      mouseCB = PyMapping_GetItemString(callbacks, "mouse");
    }
    if (PyMapping_HasKeyString(callbacks, "motion")) {
      motionCB = PyMapping_GetItemString(callbacks, "motion");
    }
  }
}

static void withdraw(void) {
  Py_XDECREF(idleCB);
  idleCB = NULL;
  Py_XDECREF(resizeCB);
  resizeCB = NULL;
  Py_XDECREF(renderCB);
  renderCB = NULL;
  Py_XDECREF(keyboardCB);
  keyboardCB = NULL;
  Py_XDECREF(specialCB);
  specialCB = NULL;
  Py_XDECREF(mouseCB);
  mouseCB = NULL;
  Py_XDECREF(motionCB);
  motionCB = NULL;
}

// Individual setup functions:

static void initContext(PyObject* args) {
  int i, argc, argvmallocd;
  char* arg = "symbiosis";
  char** argv;
  PyObject* item;

  argvmallocd = 0;

  // Set up argv to point to the given list of Bytes objects:
  argc = PyObject_Length(args);
  if (argc < 0) {
    argc = 1;
    argv = &arg;
  } else {
    argv = (char**) malloc(sizeof(char*) * argc);
    if (argv == NULL) {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }
    argvmallocd = 1;
    for (i = 0; i < argc; ++i) {
      item = PySequence_GetItem(args, i);
      if (item == NULL) {
        free(argv);
        argvmallocd = 0;
        argc = 1;
        argv = &arg;
        break;
      } else {
        if (PyBytes_Check(item)) {
          argv[i] = PyBytes_AsString(item);
        } else {
          free(argv);
          argvmallocd = 0;
          argc = 1;
          argv = &arg;
          break;
        }
        Py_DECREF(item);
      }
    }
  }

  // Initialize the GLUT context:
  glutInit(&argc, argv);

  // Free the args list if required:
  if (argvmallocd) {
    free(argv);
  }
}

static void setMode(int mode) {
  glutInitDisplayMode(mode);
}

static void setSize(int width, int height) {
  glutInitWindowSize( width, height );
}

static void createWindow(const char* name) {
  WINDOW = glutCreateWindow( name );
}

static void setBGColor(float r, float g, float b, float a) {
  glClearColor( r, g, b, a );
}

static void activateCallbacks(void) {
  // Setup the glut callback functions:
  glutIdleFunc( idle );
  glutReshapeFunc( resize );
  glutDisplayFunc( render );
  glutKeyboardFunc( keyboard );
  glutKeyboardUpFunc( keyboardUp );
  glutSpecialFunc( special );
  glutSpecialUpFunc( specialUp );
  glutMouseFunc( mouse );
  glutMotionFunc( motion );
}

// Main interface functions:

static void prepare(
  PyObject* glutargs,
  int mode,
  int w,
  int h,
  const char* name,
  float r,
  float g,
  float b,
  float a
) {
  initContext(glutargs);
  setMode(mode);
  setSize(w, h);
  createWindow(name);
  setBGColor( r, g, b, a);
  activateCallbacks();
}

static void loop(void) {
  // Swap buffers and loop away:
  glutSwapBuffers();
  glutMainLoop();
}

static void fail(void) {
  withdraw();
  glutDestroyWindow(WINDOW);
  exit(1);
}

static void quit(void) {
  withdraw();
  glutDestroyWindow(WINDOW);
  exit(0);
}

// Example rendering functions:

static void drawTest(void) {
  // Set up a fresh model view:
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Draw a blue square:
  glBegin( GL_QUADS );

  glColor3ub(32, 92, 196);

  glVertex3f(0, 0, 0);
  glVertex3f(1, 0, 0);
  glVertex3f(1, 1, 0);
  glVertex3f(0, 1, 0);

  glEnd();
}

static void clearColorBuffer(void) {
  glClear( GL_COLOR_BUFFER_BIT );
}

// Python wrapper layer functions:

static PyObject* attach_py(PyObject* self, PyObject* args) {
  PyObject* arg = NULL;

  if (!PyArg_ParseTuple(args, "O", &arg)) { return NULL; }

  attach(arg);

  Py_RETURN_NONE;
}

static PyObject* withdraw_py(PyObject* self) {
  withdraw();
  Py_RETURN_NONE;
}

static PyObject* initContext_py(PyObject* self, PyObject* args) {
  PyObject* arg = NULL;

  if (!PyArg_ParseTuple(args, "O", &arg)) { return NULL; }

  initContext(arg);

  Py_RETURN_NONE;
}

static PyObject* setMode_py(PyObject* self, PyObject* args) {
  int mode = DEFAULT_MODE;

  if (!PyArg_ParseTuple(args, "|i", &mode)) { return NULL; }

  setMode(mode);

  Py_RETURN_NONE;
}

static PyObject* setSize_py(PyObject* self, PyObject* args) {
  int w = DEFAULT_WIDTH, h = DEFAULT_HEIGHT;

  if (!PyArg_ParseTuple(args, "|ii", &w, &h)) { return NULL; }

  setSize(w, h);

  Py_RETURN_NONE;
}

static PyObject* createWindow_py(PyObject* self, PyObject* args) {
  const char* name = NULL;

  if (!PyArg_ParseTuple(args, "s", &name)) { return NULL; }

  createWindow(name);

  Py_RETURN_NONE;
}

static PyObject* setBGColor_py(PyObject* self, PyObject* args) {
  float r = 0.0, g = 0.0, b = 0.0, a = 1.0;

  if (!PyArg_ParseTuple(args, "|ffff", &r, &g, &b, &a)) { return NULL; }

  setBGColor(r, g, b, a);

  Py_RETURN_NONE;
}

static PyObject* activateCallbacks_py(PyObject* self) {
  activateCallbacks();
  Py_RETURN_NONE;
}

static PyObject* prepare_py(PyObject* self, PyObject* args, PyObject* kwargs) {
  PyObject *glutargs = NULL, *bgcolor = NULL, *item = NULL;
  int i, mode = DEFAULT_MODE, w = DEFAULT_WIDTH, h = DEFAULT_HEIGHT;
  float r = 0.0, g = 0.0, b = 0.0, a = 1.0;
  const char* name = "";
  char* kwlist[] = {
    "args",
    "mode", "width", "height",
    "name",
    "background",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(
    args,
    kwargs,
    "|OiiisO",
    kwlist,
    &glutargs,
    &mode, &w, &h,
    &name,
    &bgcolor
  )) { return NULL; }

  if (bgcolor != NULL && PySequence_Check(bgcolor)) {
    for(i = 0; i < 4; ++i) {
      item = PySequence_GetItem(bgcolor, i);
      if (item != NULL && PyNumber_Check(item)) {
        if (i == 0) {
          r = PyFloat_AsDouble(item);
        } else if (i == 1) {
          g = PyFloat_AsDouble(item);
        } else if (i == 2) {
          b = PyFloat_AsDouble(item);
        } else if (i == 3) {
          a = PyFloat_AsDouble(item);
        }
      } else {
        break;
      }
    }
  }

  if (glutargs == NULL) {
    glutargs = Py_BuildValue("[]");
  } else {
    Py_INCREF(glutargs);
  }

  prepare(glutargs, mode, w, h, name, r, g, b, a);

  Py_XDECREF(glutargs);

  Py_RETURN_NONE;
}

static PyObject* loop_py(PyObject* self) {
  loop();
  Py_RETURN_NONE;
}

static PyObject* fail_py(PyObject* self) {
  fail();
  Py_RETURN_NONE;
}

static PyObject* quit_py(PyObject* self) {
  quit();
  Py_RETURN_NONE;
}

static PyObject* drawTest_py(PyObject* self) {
  drawTest();
  Py_RETURN_NONE;
}

static PyObject* clearColorBuffer_py(PyObject* self) {
  clearColorBuffer();
  Py_RETURN_NONE;
}

// Python module definition structures:

// The Python method table:
static PyMethodDef SymbiosisMethods[] = {
  {
    "attach",
    attach_py,
    METH_VARARGS,
    "Arguments: a single dictionary containing callback functions.\n"
    "Set callbacks using a dictionary. The dictionary should contain callback "
    "functions indexed by their names, each of which accepts the right number "
    "of arguments. The callbacks only work if set_callbacks has been called. "
    "The available callback functions are:\n"
    "  idle()\n"
    "  resize(int width, int height)\n"
    "  render()\n"
    "  keyboard(bool pressed, str key, int x, int y)\n"
    "  special(bool pressed, int key, int x, int y)\n"
    "  mouse(bool pressed, int button, int x, int y)\n"
    "  motion(int x, int y)\n"
    "\n"
    "The idle, resize, and render functions include default behavior, which is "
    "triggered if either no callback is assigned, or if the assigned callback "
    "returns True. Default behaviors are as follows:\n"
    "  idle - Call glutPostRedisplay().\n"
    "  resize - Call glViewport(0, 0, width, height).\n"
    "  render - Call glutSwapBuffers(), and then clear the color buffer.\n"
  },
  {
    "withdraw",
    (PyCFunction) withdraw_py,
    METH_NOARGS,
    "Arguments: none.\n"
    "Clears all callback functions. This activates defaults for the idle, "
    "resize, and render functions (see attach)."
  },
  {
    "init_context",
    initContext_py,
    METH_VARARGS,
    "Arguments: a list of bytes objects.\n"
    "Set up OpenGL and the GLUT context using a list of bytes objects (which "
    "it gives as arguments to glutInit)."
  },
  {
    "set_mode",
    setMode_py,
    METH_VARARGS,
    "Arguments: an integer (optional) - the display mode.\n"
    "Sets the display mode, using the given integer (which should be a bitwise "
    "combination of the relevant GLUT_* constants). If no argument is given, "
    "a default of GLUT_RGBA | GLUT_DOUBLE is used."
  },
  {
    "set_size",
    setSize_py,
    METH_VARARGS,
    "Arguments: two integers (optional) - the window width and height.\n"
    "Sets the window size, using the given integers. If any arguments are "
    "missing, the defaults (800 and 600, respectively) are used in their "
    "place(s)."
  },
  {
    "create_window",
    createWindow_py,
    METH_VARARGS,
    "Arguments: a string for the window name.\n"
    "Creates a GLUT window. The init_context, set_mode, and set_size functions "
    "must have been called, or this won't work properly. This should be called "
    "before set_callbacks."
  },
  {
    "set_background",
    setBGColor_py,
    METH_VARARGS,
    "Arguments: four optional numbers between 0 and 1 - the color (RGBA).\n"
    "Sets the background color (the clear value for the color buffer). The "
    "default values of 0, 0, 0, and 1 for R, G, B, and A are used for omitted "
    "arguments. This should only be called after create_window, since it sets "
    "The clear color within the context of a window's color buffer."
  },
  {
    "activate_callbacks",
    (PyCFunction) activateCallbacks_py,
    METH_NOARGS,
    "Arguments: none.\n"
    "Activates callbacks (for the current window). It should be called after "
    "create_window. Callbacks set using attach won't work unless this is "
    "called."
  },
  {
    "prepare",
    (PyCFunction) prepare_py,
    METH_VARARGS | METH_KEYWORDS,
    "Arguments: the following optional keyword arguments:\n"
    "  args - a list of bytes objects for init_context.\n"
    "  mode - an integer for set_mode.\n"
    "  width - an integer for set_size.\n"
    "  height - an integer for set_size.\n"
    "  name - a string for create_window.\n"
    "  background - a 4-tuple of floats (an RGBA color) for set_background.\n"
    "Sets up a full GLUT context and window, filling in missing keyword "
    "arguments with defaults as necessary. Just calls all of the required "
    "setup functions in order, using the given arguments."
  },
  {
    "loop",
    (PyCFunction) loop_py,
    METH_NOARGS,
    "Arguments: none.\n"
    "Start the GLUT main loop. This function doesn't return until the GLUT "
    "main loop exits; call it only after all of the setup is complete."
  },
  {
    "quit",
    (PyCFunction) quit_py,
    METH_NOARGS,
    "Arguments: none.\n"
    "Withdraw, destroy the window, and quit the program. Using exit() from "
    "within Python won't work after calling loop, use this instead. "
    "Unfortunately there isn't any way to shut down the GLUT main loop without "
    "killing the host."
  },
  {
    "fail",
    (PyCFunction) fail_py,
    METH_NOARGS,
    "Arguments: none.\n"
    "Just like the quit function, but returns an error value (1) on "
    "termination instead of 0."
  },
  {
    "draw_test",
    (PyCFunction) drawTest_py,
    METH_NOARGS,
    "Arguments: none.\n"
    "A test function that executes some sample OpenGL drawing code. Call it "
    "from the render callback in Python."
  },
  {
    "clear_color_buffer",
    (PyCFunction) clearColorBuffer_py,
    METH_NOARGS,
    "Arguments: none.\n"
    "Clears the color buffer."
  },
  { NULL, NULL, 0, NULL } // Terminal value.
};

// The Python module definition:
static struct PyModuleDef symbiosismodule = {
  PyModuleDef_HEAD_INIT,
  "symbiosis",
"The symbiosis module provides a small bridge between GLUT/OpenGL and Python, "
"providing a Python module that includes functions for initializing a GLUT "
"context and setting up Python functions as callbacks for the standard GLUT "
"events. It doesn't provide a direct interface to OpenGL or GLUT "
"functionality, however, and it isn't meant to. The necessary calls to things "
"like OpenGL drawing primitives that must be invoked from places like a Python "
"callback function registered using symbiosis should be dispatched using "
"either ctypes calls into the relevant libraries or another application-"
"specific extension module.",
  -1,
  SymbiosisMethods
};

// The Python module initialization call:
PyMODINIT_FUNC
PyInit_symbiosis(void) {
  return PyModule_Create(&symbiosismodule);
}
